
import torch

from torch import Tensor

from escnn.group import GroupElement
from .field_type import FieldType

from typing import List, Union

import itertools
from collections.abc import Iterable


__all__ = ["GeometricTensor", "tensor_directsum"]


class GeometricTensor:
    
    def __init__(self, tensor: Tensor, type: FieldType, coords: Tensor = None):
        r"""
        
        A GeometricTensor can be interpreted as a *typed* tensor.
        It is wrapping a common :class:`torch.Tensor` and endows it with a (compatible) :class:`~escnn.nn.FieldType` as
        *transformation law*.
        
        The :class:`~escnn.nn.FieldType` describes the action of a group :math:`G` on the tensor.
        This action includes both a transformation of the base space and a transformation of the channels according to
        a :math:`G`-representation :math:`\rho`.
        
        All *escnn* neural network operations have :class:`~escnn.nn.GeometricTensor` s as inputs and outputs.
        They perform a dynamic typechecking, ensuring that the transformation laws of the data and the operation match.
        See also :class:`~escnn.nn.EquivariantModule`.
 
        As usual, the first dimension of the tensor is interpreted as the batch dimension. The second is the fiber
        (or channel) dimension, which is associated with a group representation by the field type. The following
        dimensions are the spatial dimensions (like in a conventional CNN).
        
        In addition, the method accepts an optional ``coords`` tensor.
        If the argument is not passed, the `input` tensor is assumed to have shape
        ``(batchsize, channels, *spatial_grid_shape)`` and to represent features sampled on a grid of shape
        ``spatial_grid_shape``; in this case,  the action on the base space resamples the transformed features on this
        grid (using interpolation, if necessary).
        If ``coords`` is not ``None``, ``input`` is assumed to be a ``(#points, channels)`` tensor containing an unstructured
        set of points living on the base space; then, ``coords`` should contain the coordinates of these points.
        The base space action will then transform these coordinates (no interpolation required).
        In that case, the method returns a pair containing both the transformed features (according to the action on the
        fibers) and the transformed coordinates (according to the action on the basespace).

        In addition, this class accepts an optional argument ``coords``.
        If the argument is not passed, the input ``tensor`` is assumed to have shape
        ``(batchsize, channels, *spatial_grid_shape)`` and to represent features sampled on a grid of shape
        ``spatial_grid_shape``; in this case,  the action on the base space resamples the transformed features on this
        grid (using interpolation, if necessary).
        If ``coords`` is not ``None``, ``tensor`` is assumed to be a ``(#points, channels)`` tensor containing an unstructured
        set of points living on the base space; then, `coords` should contain the coordinates of these points.
        The tensor ``coords`` must have shape ``(#points, type.gspace.dimensionality)``, where ``#points = tensor.shape[0]``.
        The base space action will then transform these coordinates (no interpolation required).
        See also the method :meth:`escnn.nn.FieldType.transform` for more details.
        
        The operations of **addition** and **scalar multiplication** are supported.
        For example::
        
            gs = escnn.gspaces.rot2dOnR2(8)
            type = escnn.nn.FieldType(gs, [gs.regular_repr]*3)
            t1 = escnn.nn.GeometricTensor(torch.randn(1, 24, 3, 3), type)
            t2 = escnn.nn.GeometricTensor(torch.randn(1, 24, 3, 3), type)
            
            # addition
            t3 = t1 + t2
            
            # scalar product
            t3 = t1 * 3.
            
            # scalar product also supports tensors containing only one scalar
            t3 = t1 * torch.tensor(3.)
            
            # inplace operations are also supported
            t1 += t2
            t2 *= 3.
        
        .. warning ::
            The multiplication of a PyTorch tensor containing only a scalar with a GeometricTensor is only supported
            when using PyTorch 1.4 or higher (see this `issue <https://github.com/pytorch/pytorch/issues/26333>`_ )
            
        .. warning ::
            These operations are only supported for *compatible* tensors, i.e. tensors which share the same ``type`` and,
            if set, ``coords``.
            
        A GeometricTensor can also be right-multiplied with a :class:`~escnn.group.GroupElement` with the ``@`` operator.
        The group element must belong to the :attr:`~escnn.nn.FieldType.fibergroup` of the :class:`~escnn.nn.FieldType`
        ``type`` of this tensor.
        The right-multiplication through ``@`` *only transforms the channels* using the group representation
        ``type.representation`` but *does not transform the base space*.
        This operation is equivalent to :meth:`~escnn.nn.GeometricTensor.transform_fibers`.
        Check its documentation for more details.
        
        
        A GeometricTensor supports **slicing** in a similar way to PyTorch's :class:`torch.Tensor`.
        More precisely, slicing along the batch (1st) and the spatial (3rd, 4th, ...) dimensions works as usual.
        However, slicing the fiber (2nd) dimension would break equivariance when splitting channels belonging to
        the same field.
        To prevent this, slicing on the second dimension is defined over *fields* instead of channels.
        Note that, if ``coords`` is not `None`, slicing over the first dimension also slices the ``coords`` tensor over its
        first dimension.
        Moreover, a GeometricTensor also partially supports **advanced indexing** (see NumPy's
        documentation about
        `indexing <https://numpy.org/doc/stable/user/basics.indexing.html#indexing-on-ndarrays>`_
        for more details).

        .. warning ::
            In contrast to NumPy and PyTorch, an index containing a single integer value **does not** reduce
            the dimensionality of the tensor.
            In this way, the resulting tensor can always be interpreted as a GeometricTensor.
            
        
        We give few examples to illustrate this behavior::
        
            # Example of GeometricTensor slicing
            space = escnn.gspaces.rot2dOnR2(4)
            type = escnn.nn.FieldType(space, [
                # field type            # index # size
                space.regular_repr,     #   0   #  4
                space.regular_repr,     #   1   #  4
                space.irrep(1),         #   2   #  2
                space.irrep(1),         #   3   #  2
                space.trivial_repr,     #   4   #  1
                space.trivial_repr,     #   5   #  1
                space.trivial_repr,     #   6   #  1
            ])                          #   sum = 15
            
            # this FieldType contains 7 fields
            len(type)
            >> 7
            
            # the size of this FieldType is equal to the sum of the sizes of each of its fields
            type.size
            >> 15
            
            geom_tensor = escnn.nn.GeometricTensor(torch.randn(10, type.size, 9, 9), type)

            # or, equivalently:
            geom_tensor = type(torch.randn(10, type.size, 9, 9))

            geom_tensor.shape
            >> torch.Size([10, 15, 9, 9])
            
            geom_tensor[1:3, :, 2:5, 2:5].shape
            >> torch.Size([2, 15, 3, 3])
            
            geom_tensor[..., 2:5].shape
            >> torch.Size([10, 15, 9, 3])
            
            # the tensor contains the fields 1:4, i.e 1, 2 and 3
            # these fields have size, respectively, 4, 2 and 2
            # so the resulting tensor has 8 channels
            geom_tensor[:, 1:4, ...].shape
            >> torch.Size([10, 8, 9, 9])
            
            # the tensor contains the fields 0:6:2, i.e 0, 2 and 4
            # these fields have size, respectively, 4, 2 and 1
            # so the resulting tensor has 7 channels
            geom_tensor[:, 0:6:2].shape
            >> torch.Size([10, 7, 9, 9])
            
            # the tensor contains only the field 2, which has size 2
            # note, also, that even though a single index is used for the batch dimension, the resulting tensor
            # still has 4 dimensions
            geom_tensor[3, 2].shape
            >> torch.Size(1, 2, 9, 9)

            # we can use a boolean tensor to perform advanced indexing over the fields of a Geometric Tensor.
            # the tensor contains only the two fields of size 2, i.e. the third and the fourth.
            # moreover, we also slice the batch dimension.
            idx = torch.tensor([type.representations[i].size == 2 for i in range(len(type))])
            geom_tensor[0:4, idx].shape
            >> torch.Size(4, 4, 9, 9)

            # it is also possible to use an integer tensor to perform advanced indexing.
            # Here, we select the second and the sixth fields.
            idx = torch.tensor([1, 5])
            geom_tensor[0:4, idx].shape
            >> torch.Size(4, 5, 9, 9)

            # advanced indexing over the other dimensions works as usual.
            idx = torch.tensor([1, 2, 2, 8, 4])
            geom_tensor[idx, 1:3].shape
            >> torch.Size(5, 6, 9, 9)

        
        .. warning ::
        
            *Slicing* over the fiber (2nd) dimension with ``step > 1`` or with a negative step is converted
            into *indexing* over the channels.
            This means that, in these cases, slicing behaves like *advanced indexing* in PyTorch and NumPy
            **returning a copy instead of a view**.
            For more details, see the *note* `here <https://pytorch.org/docs/stable/tensor_view.html>`_ and
            *NumPy*'s `docs <https://docs.scipy.org/doc/numpy/reference/arrays.indexing.html>`_ .
        
        .. note ::
    
            Slicing is not supported for setting values inside the tensor
            (i.e. :meth:`~object.__setitem__` is not implemented).
            Indeed, depending on the values which are assigned, this operation can break the symmetry of the tensor
            which may not transform anymore according to its transformation law (specified by ``type``).
            In case this feature is necessary, one can directly access the underlying :class:`torch.Tensor`, e.g.
            ``geom_tensor.tensor[:3, :, 2:5, 2:5] = torch.randn(3, 4, 3, 3)``, although this is not recommended.

        .. warning ::
            Using advanced indexing over multiple axes simultaneously drops the corresponding dimensions (except for the
            first one). This means that the resulting tensor will likely not have a shape compatible with a Geometric
            Tensor (e.g. the spatial dimensions are dropped), causing an error.
            
            
        Args:
            tensor (torch.Tensor): the tensor data
            type (FieldType): the type of the tensor, modeling its transformation law
        
        Attributes:
            ~.tensor (torch.Tensor)
            ~.type (FieldType)
            
        """
        assert isinstance(tensor, torch.Tensor)
        assert isinstance(type, FieldType)
        
        if coords is None:
            assert len(tensor.shape) == 2 + type.gspace.dimensionality, (tensor.shape, type.gspace.dimensionality)
        else:
            assert len(coords.shape) == 2, coords.shape

            assert coords.shape[1] == type.gspace.dimensionality, \
                f"Error! `coords` tensor with shape {coords.shape} does not match the dimensionality of the field type {type.gspace.dimensionality}."

            assert len(tensor.shape) == 2, \
                f"Error! If `coords` is not `None`, the tensor must be a 2D tensor of shape (#points, filedtype size) but the tensor has shape {tensor.shape}."

            assert tensor.shape[0] == coords.shape[0],\
                f"Error! Points in the `tensor` and `coords` tensors do not match: {tensor.shape[0]} != {coords.shape[0]}."

        assert tensor.shape[1] == type.size, \
            f"Error! The size of the tensor {tensor.shape} does not match the size of the field type {type.size}."

        # torch.Tensor: PyTorch tensor containing the data
        self.tensor = tensor

        # torch.Tensor: PyTorch tensor containing the coordinates of the data points
        self.coords = coords

        # FieldType: field type of the signal
        self.type = type
    
    def restrict(self, id) -> 'GeometricTensor':
        r"""
        Restrict the field type of this tensor.
        
        The method returns a new :class:`~escnn.nn.GeometricTensor` whose :attr:`~escnn.nn.GeometricTensor.type`
        is equal to this tensor's :attr:`~escnn.nn.GeometricTensor.type`
        restricted to a subgroup :math:`H<G` (see :meth:`escnn.nn.FieldType.restrict`).
        The restricted :attr:`~escnn.nn.GeometricTensor.type` is associated with the restricted representation
        :math:`\Res{H}{G}\rho` of the :math:`G`-representation :math:`\rho` associated to this tensor's
        :attr:`~escnn.nn.GeometricTensor.type`.
        The input ``id`` specifies the subgroup :math:`H < G`.
        
        Notice that the underlying :attr:`~escnn.nn.GeometricTensor.tensor` instance will be shared between
        the current tensor and the returned one.
        
        .. warning ::
        
            The method builds the new representation on the fly; hence, if this operation is needed at run time,
            we suggest to use :class:`escnn.nn.RestrictionModule` which pre-computes the new representation offline.
        
        .. seealso ::
        
            Check the documentation of the :meth:`~escnn.gspaces.GSpace.restrict` method in the
            :class:`~escnn.gspaces.GSpace` instance used for a description of the parameter ``id``.
        
        
        Args:
            id: the id identifying the subgroup :math:`H` the representations are restricted to

        Returns:
            the geometric tensor with the restricted representations
            
        """
        new_class = self.type.restrict(id)
        return GeometricTensor(self.tensor, new_class, self.coords)
    
    def split(self, breaks: List[int]):
        r"""
        
        Split this tensor on the channel dimension in a list of smaller tensors.
        The original tensor is split at the *fields* specified by the index list ``breaks``.
        
        If the tensor is associated with the list of fields :math:`\{\rho_i\}_i`
        (see :attr:`escnn.nn.FieldType.representations`), the :math:`j`-th output tensor will contain the fields
        :math:`\rho_{\text{breaks}[j-1]}, \dots, \rho_{\text{breaks}[j]-1}` of the original tensor.
        
        If `breaks = None`, the tensor is split at each field.
        This is equivalent to using `breaks = list(range(len(self.type)))`.
        
        Example ::
        
            space = escnn.gspaces.rot2dOnR2(4)
            type = escnn.nn.FieldType(space, [
                space.regular_repr,     # size = 4
                space.regular_repr,     # size = 4
                space.irrep(1),         # size = 2
                space.irrep(1),         # size = 2
                space.trivial_repr,     # size = 1
                space.trivial_repr,     # size = 1
                space.trivial_repr,     # size = 1
            ])                          #  sum = 15
            
            type.size
            >> 15
            
            geom_tensor = escnn.nn.GeometricTensor(torch.randn(10, type.size, 7, 7), type)
            
            geom_tensor.shape
            >> torch.Size([10, 15, 7, 7])
            
            # split the tensor in 3 parts
            len(geom_tensor.split([0, 4, 6]))
            >> 3
            
            # the first contains
            # - the first 2 regular fields (2*4 = 8 channels)
            # - 2 vector fields (irrep(1)) (2*2 = 4 channels)
            # and, therefore, contains 12 channels
            geom_tensor.split([0, 4, 6])[0].shape
            >> torch.Size([10, 12, 7, 7])
            
            # the second contains only 2 scalar (trivial) fields (2*1 = 2 channels)
            geom_tensor.split([0, 4, 6])[1].shape
            >> torch.Size([10, 2, 7, 7])
            
            # the last contains only 1 scalar (trivial) field (1*1 = 1 channels)
            geom_tensor.split([0, 4, 6])[2].shape
            >> torch.Size([10, 1, 7, 7])
            
        
        Args:
            breaks (list): indices of the fields where to split the tensor

        Returns:
            list of :class:`~escnn.nn.GeometricTensor` s into which the original tensor is chunked
            
        """
        if breaks is None:
            breaks = list(range(len(self.type)))
            
        breaks.append(len(self.type))
        
        # final list of tensors
        tensors = []
        
        # list containing the index of the channels separating consecutive fields in this tensor
        positions = []
        last = 0
        for repr in self.type.representations:
            positions.append(last)
            last += repr.size
        positions.append(last)
        
        last_field = 0
        # for each break point
        for b in breaks:
            assert b > last_field, 'Error! "breaks" must be an increasing list of positive indexes'
            
            # compute the sub-class of the new sub-tensor
            repr = FieldType(self.type.gspace, self.type.representations[last_field: b])
            
            # retrieve the sub-tensor
            data = self.tensor[:, positions[last_field]:positions[b], ...]
            
            tensors.append(GeometricTensor(data, repr, self.coords))
            
            last_field = b
        
        return tensors

    def transform(self, element: GroupElement) -> 'GeometricTensor':
        r"""
        Transform the current tensor according to the group representation associated to the input element
        and its induced action on the base space

        .. warning ::
            The input tensor is detached before the transformation therefore no gradient is backpropagated
            through this operation

            See :meth:`escnn.nn.GeometricTensor.transform_fibers` to transform only the fibers, i.e. not transform
            the base space.

        Args:
            element (GroupElement): an element of the group of symmetries of the fiber.

        Returns:
            the transformed tensor

        """

        if self.coords is None:
            transformed = self.type.transform(self.tensor, element, coords=None)
            coords = None
        else:
            transformed, coords = self.type.transform(self.tensor, element, self.coords)

        return GeometricTensor(transformed, self.type, coords)

    def transform_fibers(self, element: GroupElement) -> 'GeometricTensor':
        r"""
        
        Transform the feature vectors of the underlying tensor according to the group representation associated to
        the input element.
        
        Interpreting the tensor as a vector-valued signal :math:`f: X \to \mathbb{R}^c` over a base space :math:`X`
        (where :math:`c` is the number of channels of the tensor), given the input ``element`` :math:`g \in G`
        (:math:`G` fiber group) the method returns the new signal :math:`f'`:
        
        .. math ::
            f'(x) := \rho(g) f(x)
        
        for :math:`x \in X` point in the base space and :math:`\rho` the representation of :math:`G` in the
        field type of this tensor.
        
        
        Notice that the input element has to be an element of the fiber group of this tensor's field type.
        
        .. seealso ::
        
            See :meth:`escnn.nn.GeometricTensor.transform` to transform the whole tensor.
        
        Args:
            element (GroupElement): an element of the group of symmetries of the fiber.

        Returns:
            the transformed tensor

        """
        
        transformed_tensor = self.type.transform_fibers(self.tensor, element)
        return GeometricTensor(transformed_tensor, self.type, self.coords)
    
    @property
    def shape(self):
        r"""
        Alias for ``self.tensor.shape``

        """
        return self.tensor.shape

    def size(self):
        r"""
        Alias for ``self.tensor.size()``

        .. seealso ::
            :meth:`torch.Tensor.size`

        """
        return self.tensor.size()

    def to(self, *args, **kwargs):
        r"""
        
        Alias for ``self.tensor.to(*args, **kwargs)``.
        
        Applies :meth:`torch.Tensor.to` to the underlying tensor and wraps the resulting tensor in a new
        :class:`~escnn.nn.GeometricTensor` with the same type.
        
        .. warning ::
            This method does not affect ``self.coords``.

        """
        tensor = self.tensor.to(*args, **kwargs)
        return GeometricTensor(tensor, self.type, self.coords)

    def __getitem__(self, slices) -> 'GeometricTensor':
        r'''
        
        A GeometricTensor supports *slicing* in a similar way to PyTorch's :class:`torch.Tensor`.
        More precisely, slicing along the batch (1st) and the spatial (3rd, 4th, ...) dimensions works as usual.
        However, slicing along the channel dimension could break equivariance by splitting the channels belonging to the
        same field.
        For this reason, slicing on the second dimension is not defined over the channels but over fields.
        
        When a continuous (step=1) slice is used over the fields/channels dimension (the 2nd axis), it is converted
        into a continuous slice over the channels.
        This is not possible when the step is greater than 1 or negative.
        In such cases, the slice over the fields needs to be converted into an index over the channels.
        
        Moreover, when a single integer is used to index an axis, that axis is not discarded as in PyTorch but is
        preserved with size 1.

        If slicing is performed over the batch dimension and `coords` is not `None`, also the `coords` tensor is sliced.
        
        Slicing is not supported for setting values inside the tensor (i.e. :meth:`object.__setitem__`
        is not implemented).
        
        '''
        
        # Slicing is not supported on the channel dimension.
        if isinstance(slices, tuple):
            if len(slices) > len(self.tensor.shape):
                raise TypeError(
                    f'''
                        Error! Too many slicing indices for GeometricTensor.
                        At most {len(self.tensor.shape)} indices expected but {len(slices)} indices passed.
                    '''
                )
        else:
            slices = (slices,)
            
        for i, idx in enumerate(slices):
            if (isinstance(idx, slice) or isinstance(idx, int) or idx == Ellipsis):
                continue
            elif isinstance(idx, torch.Tensor) and idx.dtype == torch.bool:
                continue
            elif isinstance(idx, torch.Tensor) and idx.dtype == torch.long:
                continue
            else:
                raise TypeError(f'''
                        Error! Advanced Indexing over a GeometricTensor is not fully supported yet.
                        Currently, only indexing with boolean or long tensors and basic slicing are supported.
                ''')
        
        naxes = len(self.tensor.shape)
        
        # count the number of indexes passed
        indexed_axes = 0
        for idx in slices:
            indexed_axes += 1 - (idx == Ellipsis)
        
        # number of axes which are missing an index
        missing_axes = naxes - indexed_axes
        
        # expand the first ellipsis with a number of full slices (i.e. [::]) equal to the number
        # of axes not indexed. Discard all other ellipses
        expanded_idxs = []
        expanded_ellipsis = False
        for s in slices:
            if s == Ellipsis:
                # expand only the first ellipsis
                if not expanded_ellipsis:
                    expanded_idxs += [slice(None)]*missing_axes
                    expanded_ellipsis = True
            else:
                # other indices are preserved
                expanded_idxs.append(s)
        
        # maximum index per dimension
        idx_max = list(self.tensor.shape)
        idx_max[1] = len(self.type)
        
        # If an index containing a single integer is passed, it is converted into a slice
        # which starts at that index and ends at the following one.
        # In this way, when passing a single integer to index a dimension, the resulting tensor will still have that
        # dimension with size 1
        for i in range(len(expanded_idxs)):
            if isinstance(expanded_idxs[i], int):
                idx = expanded_idxs[i]
                if idx < 0:
                    # convert a negative index into a positive index
                    idx = idx_max[i] + idx
                expanded_idxs[i] = slice(idx, idx+1, 1)
        
        if len(expanded_idxs) == 1:
            # if only the first dimension is indexed, there is no need to do anything
            # the resulting tensor will have the same type of the original as the indexing does not affect the
            # channels/fields dimension
            type = self.type
            
        elif isinstance(expanded_idxs[1], slice) and (expanded_idxs[1].step is None or expanded_idxs[1].step == 1):
            # If the index over the fields is a slice and it is contiguous, we can convert it into a
            # contiguous slice over the channels.
            # The slice will start from the first channel of the first field and will stop at the last channel
            # of the last field
            start = expanded_idxs[1].start if expanded_idxs[1].start is not None else 0
            stop = expanded_idxs[1].stop if expanded_idxs[1].stop is not None else len(self.type)
            channel_idxs = slice(
                self.type.fields_start[start],
                self.type.fields_end[stop-1],
                1
            )
            
            if start == 0 and stop == len(self.type):
                # if all the fields are retrieved by this index, the resulting tensor has the same field
                # types of the original one
                type = self.type
            else:
                # otherwise, only a subset of the fields are preserved
                type = FieldType(self.type.gspace, self.type.representations[expanded_idxs[1]])
                
            expanded_idxs[1] = channel_idxs

        else:
            # If the index over the fields is not a slice or it is not a contiguous slice, we need to convert it
            # into an index over the channels. We first use the index provided to retrieve the list of fields
            # and then add the index of their channels in a list of indexes
            idxs = []

            # convert the indices into iterable and retrieve the subset of field representations
            if isinstance(expanded_idxs[1], slice):
                fields = range(len(self.type))[expanded_idxs[1]]
                representations = self.type.representations[expanded_idxs[1]]
            elif isinstance(expanded_idxs[1], int):
                fields = [expanded_idxs[1]]
                representations = self.type.representations[expanded_idxs[1]]
            elif isinstance(expanded_idxs[1], torch.Tensor) and expanded_idxs[1].dtype == torch.bool:
                fields = [i for i in range(len(self.type)) if expanded_idxs[1][i]]
                representations = [self.type.representations[f] for f in fields]
            elif isinstance(expanded_idxs[1], torch.Tensor) and expanded_idxs[1].dtype == torch.long:
                fields = expanded_idxs[1].tolist()
                representations = [self.type.representations[f] for f in fields]
            elif isinstance(expanded_idxs[1], Iterable):
                fields = expanded_idxs[1]
                representations = [self.type.representations[f] for f in fields]
            else:
                raise ValueError('Index over the fiber (2nd) dimension not recognized.')

            # iterate over all fields indexed by the user
            for field in fields:
                # append the indexes of the channels in the field
                idxs.append(list(
                    range(
                        self.type.fields_start[field],
                        self.type.fields_end[field],
                        1
                    )
                ))

            # only a subset of the fields are preserved by this index
            type = FieldType(self.type.gspace, representations)
            
            # concatenate all the channel indexes
            channel_idxs = list(itertools.chain(*idxs))
            expanded_idxs[1] = channel_idxs

        idxs = tuple(expanded_idxs)
        
        sliced_tensor = self.tensor[idxs]

        if self.coords is not None:
            coords = self.coords[idxs[0], :]
        else:
            coords = None

        return GeometricTensor(sliced_tensor, type, coords)
    
    def has_same_coords(self, other: 'GeometricTensor') -> bool:
        if self.coords is other.coords:
            # this includes also the case both are None
            return True
        
        if (self.coords is None) != (other.coords is None):
            return False
        elif self.coords is not None:
            return torch.allclose(self.coords, other.coords)
        return True

    def is_compatible(self, other: 'GeometricTensor') -> bool:
        if self.type != other.type:
            return False
        return self.has_same_coords(other)

    def __add__(self, other: 'GeometricTensor') -> 'GeometricTensor':
        r"""
        Add two compatible :class:`~escnn.nn.GeometricTensor` using pointwise addition.
        The two tensors needs to have the same shape and be associated to the same field type.

        Args:
            other (GeometricTensor): the other geometric tensor

        Returns:
            the sum

        """
        if isinstance(other, GeometricTensor):
            assert self.is_compatible(other), 'The two geometric tensor must have the same `type` and `coords`'
            return GeometricTensor(self.tensor + other.tensor, self.type, self.coords)
        else:
            return NotImplemented

    def __sub__(self, other: 'GeometricTensor') -> 'GeometricTensor':
        r"""
        Subtract two compatible :class:`~escnn.nn.GeometricTensor` using pointwise subtraction.
        The two tensors needs to have the same shape and be associated to the same field type.

        Args:
            other (GeometricTensor): the other geometric tensor

        Returns:
            their difference

        """
        if isinstance(other, GeometricTensor):
            assert self.is_compatible(other), 'The two geometric tensor must have the same `type` and `coords`'
            return GeometricTensor(self.tensor - other.tensor, self.type, self.coords)
        else:
            return NotImplemented

    def __iadd__(self, other: 'GeometricTensor') -> 'GeometricTensor':
        r"""
        Add a compatible :class:`~escnn.nn.GeometricTensor` to this tensor inplace.
        The two tensors needs to have the same shape and be associated to the same field type.

        Args:
            other (GeometricTensor): the other geometric tensor

        Returns:
            this tensor

        """
        if isinstance(other, GeometricTensor):
            assert self.is_compatible(other), 'The two geometric tensor must have the same `type` and `coords`'
            self.tensor += other.tensor
            return self
        else:
            return NotImplemented

    def __isub__(self, other: 'GeometricTensor') -> 'GeometricTensor':
        r"""
        Subtract a compatible :class:`~escnn.nn.GeometricTensor` to this tensor inplace.
        The two tensors needs to have the same shape and be associated to the same field type.

        Args:
            other (GeometricTensor): the other geometric tensor

        Returns:
            this tensor

        """
        if isinstance(other, GeometricTensor):
            assert self.is_compatible(other), 'The two geometric tensor must have the same `type` and `coords`'
            self.tensor -= other.tensor
            return self
        else:
            return NotImplemented

    def __mul__(self, other: Union[float, torch.Tensor]) -> 'GeometricTensor':
        r"""
        Scalar product of this :class:`~escnn.nn.GeometricTensor` with a scalar.
        
        The scalar can be a float number or a :class:`torch.Tensor` containing only
        one scalar (i.e. :func:`torch.numel` should return `1`).

        Args:
            other : a scalar

        Returns:
            the scalar product

        """
        if isinstance(other, float) or (isinstance(other, torch.Tensor) and other.numel() == 1):
            # Only multiplication with a scalar is allowed
            return GeometricTensor(self.tensor * other, self.type, self.coords)
        else:
            return NotImplemented

    __rmul__ = __mul__

    def __imul__(self, other: Union[float, torch.Tensor]) -> 'GeometricTensor':
        r"""
        Scalar product of this :class:`~escnn.nn.GeometricTensor` with a scalar.
        The operation is done inplace.

        The scalar can be a float number of a :class:`torch.Tensor` containing only
        one scalar (i.e. :func:`torch.numel` should return `1`).

        Args:
            other : a scalar

        Returns:
            the scalar product

        """
        if isinstance(other, float) or (isinstance(other, torch.Tensor) and other.numel() == 1):
            self.tensor *= other
            return self
        else:
    
            return NotImplemented

    def __rmatmul__(self, other: GroupElement):
        r"""
        
        Equivalent to :meth:`escnn.nn.GeometricTensor.transform_fibers`.
        
        .. warning::
            This only transforms the fibers, not the basespace.
        
        .. todo ::
            should this be like `.transform(element)` ?

        """
        
        if isinstance(other, GroupElement):
            assert other.group == self.type.fibergroup
            return GeometricTensor(
                self.type.transform_fibers(self.tensor, other),
                self.type,
                self.coords
            )
        else:
            return NotImplemented

    def __repr__(self):
        t = repr(self.tensor)[:-1]
        t = t.replace('\n', '\n  ')
        r = 'g_' + t + ', ' + repr(self.type) + ')'

        return r


def tensor_directsum(tensors: List['GeometricTensor']) -> 'GeometricTensor':
    r"""
    Concatenate a list of :class:`~escnn.nn.GeometricTensor` s on the channels dimension (``dim=1``).
    The input tensors have to be compatible: they need to have the same shape except for the channels
    dimension (``dim=1``); additionally, if their `coords` attribute is not `None`, they also need to share the same
    value.
    
    In the resulting :class:`~escnn.nn.GeometricTensor`, the channels dimension will be associated with the direct sum
    representation of the representations of the input tensors.
    
    .. seealso::
        
        :func:`escnn.group.directsum`
    
    Args:
        tensors (list): a list of :class:`~escnn.nn.GeometricTensor` s

    Returns:
        the direct sum of the inputs
        
    """
    # assert len(tensors) > 1
    
    for i in range(1, len(tensors)):
        assert tensors[0].type.gspace == tensors[i].type.gspace
        assert tensors[0].tensor.ndimension() == tensors[i].tensor.ndimension()
        assert tensors[0].tensor.shape[0] == tensors[i].tensor.shape[0]
        assert tensors[0].tensor.shape[2:] == tensors[i].tensor.shape[2:]
        assert tensors[0].has_same_coords(tensors[i])
    
    # concatenate all representations from all field types
    reprs = []
    for t in tensors:
        reprs += t.type.representations
    
    # build the new field type
    cls = FieldType(tensors[0].type.gspace, reprs)
    
    # concatenate the underlying tensors
    data = torch.cat([t.tensor for t in tensors], dim=1)
    
    # build the new Geometric Tensor
    return GeometricTensor(data, cls, tensors[0].coords)






import torch
import numpy as np

from .equivariant_module import EquivariantModule
from escnn.nn import GeometricTensor
from escnn.nn import FieldType
from escnn.gspaces import *

import torch
from typing import List, Tuple, Any

__all__ = ["RestrictionModule"]


class RestrictionModule(EquivariantModule):

    def __init__(self, in_type: FieldType, id):
        r"""
        
        Restricts the type of the input to the subgroup identified by ``id``.
        
        It computes the output type in the constructor and wraps the underlying tensor (:class:`torch.Tensor`) in input
        with the output type during the forward pass.
        
        This module only acts as a wrapper for :meth:`escnn.nn.FieldType.restrict`
        (or :meth:`escnn.nn.GeometricTensor.restrict`).
        The accepted values of ``id`` depend on the underlying ``gspace`` in the input type ``in_type``; check the
        documentation of the method :meth:`escnn.gspaces.GSpace.restrict` of the gspace used for
        further information.
        
        
        .. seealso::
            :meth:`escnn.nn.FieldType.restrict`,
            :meth:`escnn.nn.GeometricTensor.restrict`,
            :meth:`escnn.gspaces.GSpace.restrict`
        
        Args:
            in_type (FieldType): the input field type
            id: a valid id for a subgroup of the space associated with the input type
            
        """
        assert isinstance(in_type, FieldType)
        assert isinstance(in_type.gspace, GSpace)
        
        super(EquivariantModule, self).__init__()
        
        self._id = id
        self.in_type = in_type
        self.out_type = in_type.restrict(id)

    def forward(self, input: GeometricTensor) -> GeometricTensor:
        assert input.type == self.in_type
        return GeometricTensor(input.tensor, self.out_type, input.coords)
    
    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:
        return input_shape
    
    def check_equivariance(self, atol: float = 1e-7, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
        
        _, parent_mapping, _ = self.in_type.gspace.restrict(self._id)
        
        c = self.in_type.size
    
        x = torch.randn(3, c, 10, 10)
    
        x = GeometricTensor(x, self.in_type)
    
        errors = []
    
        for el in self.out_type.testing_elements:
            print(el)
        
            out1 = self(x).transform(el).tensor.detach().numpy()
            out2 = self(x.transform(parent_mapping(el))).tensor.detach().numpy()
        
            errs = out1 - out2
            errs = np.abs(errs).reshape(-1)
            print(el, errs.max(), errs.mean(), errs.var())
        
            assert np.allclose(out1, out2, atol=atol, rtol=rtol), \
                'The error found during equivariance check with element "{}" is too high: max = {}, mean = {} var ={}' \
                    .format(el, errs.max(), errs.mean(), errs.var())
        
            errors.append((el, errs.mean()))
    
        return errors

    def export(self):
        r"""
        Export this module to a normal PyTorch :class:`torch.nn.Identity` module and set to "eval" mode.

        .. warning ::
            Only working with PyTorch >= 1.2

        """
        self.eval()
        return torch.nn.Identity()






from .norm import NormNonLinearity
from .induced_norm import InducedNormNonLinearity
from .pointwise import PointwiseNonLinearity
from .concatenated import ConcatenatedNonLinearity
from .gated1 import GatedNonLinearity1, GATES_ID, GATED_ID
from .gated2 import GatedNonLinearity2
from .gated3 import GatedNonLinearityUniform
from .induced_gated1 import InducedGatedNonLinearity1
from .vectorfield import VectorFieldNonLinearity

from .relu import ReLU
from .elu import ELU

from .fourier import *
from .fourier_quotient import *

from .tensor import *

__all__ = [
    "NormNonLinearity",
    "InducedNormNonLinearity",
    "PointwiseNonLinearity",
    "ConcatenatedNonLinearity",
    "GatedNonLinearity1",
    "GatedNonLinearity2",
    "GatedNonLinearityUniform",
    "InducedGatedNonLinearity1",
    "VectorFieldNonLinearity",
    "ReLU",
    "ELU",
    "FourierPointwise",
    "FourierELU",
    "QuotientFourierPointwise",
    "QuotientFourierELU",
    "TensorProductModule",
]






import numpy as np
from abc import ABC, abstractmethod
from typing import List, Union, Tuple

import torch



class EmptyBasisException(Exception):
    def __init__(self):
        r"""
        Exception raised when a :class:`~escnn.kernels.KernelBasis` with no elements is built.
        
        """
        message = "The KernelBasis you tried to instantiate is empty (dim = 0). You should catch this exception."
        super(EmptyBasisException, self).__init__(message)
        

class KernelBasis(torch.nn.Module, ABC):
    
    def __init__(self, dim: int, shape: Tuple[int, int]):
        r"""
        
        Abstract class for implementing the basis of a kernel space.
        A kernel space is the space of functions in the form:
        
        .. math::
            \mathcal{K} := \{ \kappa: X \to \mathbb{R}^{c_\text{out} \times c_\text{in}} \}
        
        where :math:`X` is the base space on which the kernel is defined.
        For instance, for planar images :math:`X = \R^2`.

        One can also access the dimensionality ``dim`` of this basis via the ``len()`` method.
        
        Args:
            dim (int): the dimensionality of the basis :math:`|\mathcal{K}|` (number of elements)
            shape (tuple): a tuple containing :math:`c_\text{out}` and :math:`c_\text{in}`
            
        Attributes:
            ~.dim (int): the dimensionality of the basis :math:`|\mathcal{K}|` (number of elements)
            ~.shape (tuple): a tuple containing :math:`c_\text{out}` and :math:`c_\text{in}`
            
        """
        assert isinstance(dim, int), (dim, type(dim))
        assert isinstance(shape, tuple) and len(shape) == 2, shape
        
        assert dim >= 0
        
        if dim == 0:
            raise EmptyBasisException()
        
        self.dim = dim
        self.shape = shape

        super(KernelBasis, self).__init__()

    @abstractmethod
    def sample(self, points: torch.Tensor, out: torch.Tensor = None) -> torch.Tensor:
        r"""
        Sample the continuous basis elements on discrete points in ``points``.
        Optionally, store the resulting multidimentional array in ``out``.

        ``points`` must be an array of shape `(N, D)`, where `D` is the dimensionality of the (parametrization of the)
        base space while `N` is the number of points.

        Args:
            points (~torch.Tensor): points where to evaluate the basis elements
            out (~torch.Tensor, optional): pre-existing array to use to store the output

        Returns:
            the sampled basis

        """
        pass

    def forward(self, points: torch.Tensor, out: torch.Tensor = None) -> torch.Tensor:
        r"""
            Alias for :meth:`~escnn.kernels.KernelBasis.sample`.
        """
        return self.sample(points, out=out)

    def __len__(self):
        return self.dim

    def __iter__(self):
        for i in range(self.dim):
            yield self[i]

    @abstractmethod
    def __getitem__(self, idx: int) -> dict:
        pass
    
    @abstractmethod
    def __hash__(self):
        pass
    
    @abstractmethod
    def __eq__(self, other):
        pass


class AdjointBasis(KernelBasis):
    
    def __init__(self, basis: KernelBasis, adjoint: np.ndarray):
        r"""

        Transform the input ``basis`` by applying a change of basis ``adjoint`` on the points before sampling the basis.

        Args:
            basis (KernelBasis): a kernel basis
            adjoint (~numpy.ndarray): an orthonormal matrix defining the change of basis on the base space

        """

        n = adjoint.shape[0]
        assert adjoint.shape == (n, n)
        
        # adjoint_inv = np.linalg.inv(adjoint)
        assert np.allclose(adjoint @ adjoint.T, np.eye(n)), 'Error! The adjunction matrix must be orthonormal'
        assert np.allclose(adjoint.T @ adjoint, np.eye(n)), 'Error! The adjunction matrix must be orthonormal'

        super(AdjointBasis, self).__init__(basis.dim, basis.shape)
        
        self.basis = basis

        self.register_buffer('adj', torch.tensor(adjoint, dtype=torch.float32))
    
    def sample(self, points: torch.Tensor, out: torch.Tensor = None) -> torch.Tensor:
        r"""

        Sample the continuous basis elements on the discrete set of points.
        Optionally, store the resulting multidimentional array in ``out``.

        ``points`` must be an array of shape `(N, d)`, where `N` is the number of points and `d` their
        dimensionality.

        Args:
            points (~numpy.ndarray): points where to evaluate the basis elements
            out (~numpy.ndarray, optional): pre-existing array to use to store the output

        Returns:
            the sampled basis

        """
        assert len(points.shape) == 2
        assert points.shape[1] == self.adj.shape[0]
        
        transformed_points = points @ self.adj.to(device=points.device, dtype=points.dtype).T
        return self.basis.sample(transformed_points, out)
    
    def __getitem__(self, r):
        return self.basis[r]
    
    def __eq__(self, other):
        if isinstance(other, AdjointBasis):
            return self.basis == other.basis and torch.allclose(self.adj, other.adj)
        # elif self.basis == other:
        #     return np.allclose(self.adj, np.eye(self.adj.shape[0))
        else:
            return False
    
    def __hash__(self):
        return hash(self.adj) + 1000 * hash(self.basis)


class UnionBasis(KernelBasis):

    def __init__(self, bases_list: List[KernelBasis]):
        r"""
        Construct the union of a list of bases.
        All bases must have the same ``shape``; the resulting basis has ``dim`` equal to the sum of the dimensionalities
        of the individual bases.

        """

        if len(bases_list) == 0:
            raise EmptyBasisException

        shape = bases_list[0].shape
        dim = 0
        for i in range(len(bases_list)):
            assert bases_list[i].shape == shape
            dim += bases_list[i].dim

        if dim == 0:
            raise EmptyBasisException

        super(UnionBasis, self).__init__(dim, shape)
        self._bases = torch.nn.ModuleList(bases_list)

    def sample(self, points: torch.Tensor, out: torch.Tensor = None) -> torch.Tensor:

        assert len(points.shape) == 2
        S = points.shape[0]

        if out is None:
            out = torch.empty(S, self.dim, self.shape[0], self.shape[1], device=points.device, dtype=points.dtype)

        p = 0
        for i in range(len(self._bases)):
            basis = self._bases[i]

            basis.sample(points, out=out[:, p:p+basis.dim, ...])

            p += basis.dim

        return out

    def __getitem__(self, idx: int) -> dict:
        assert idx < self.dim
        p = 0
        for i in range(len(self._bases)):
            basis = self._bases[i]

            if idx < p + basis.dim:
                break

            p += basis.dim

        attr = self._bases[i][idx - p]

        attr["shape"] = self.shape
        attr['basis_id'] = i
        attr['basis_idx'] = attr['idx']
        attr['idx'] = idx

        return attr

    def __hash__(self):
        return sum(hash(self._bases[i])*(i+1)**2 for i in range(len(self._bases)))

    def __eq__(self, other):
        if not isinstance(other, UnionBasis):
            return False
        elif self.dim != other.dim or self.shape != other.shape or len(self._bases) != len(other._bases):
            return False
        else:
            for i in range(len(self._bases)):
                if self._bases[i] != other._bases[i]:
                    return False
            return True

    def __iter__(self):
        idx = 0
        for i, basis in enumerate(self._bases):
            for attr in basis:
                attr["shape"] = self.shape
                attr['basis_id'] = i
                attr['basis_idx'] = attr['idx']
                attr['idx'] = idx

                assert idx < self.dim

                yield attr

                idx += 1








from escnn.nn.modules.basismanager import BasisManager

from collections import defaultdict

import torch
from scipy import stats
import math

__all__ = ["generalized_he_init", "deltaorthonormal_init"]




def _generalized_he_init_variances(basismanager: BasisManager):
    r"""

    Compute the variances of the weights of a convolutional layer with a generalized He's weight initialization method.

    Args:
        basismanager (BasisManager): the basis expansion method

    """

    vars = torch.ones((basismanager.dimension(),))

    inputs_count = defaultdict(lambda: set())
    basis_count = defaultdict(int)

    basis_info = list(basismanager.get_basis_info())

    for attr in basis_info:
        i, o = attr["in_irreps_position"], attr["out_irreps_position"]
        in_irrep, out_irrep = attr["in_irrep"], attr["out_irrep"]
        inputs_count[o].add(in_irrep)
        basis_count[(in_irrep, o)] += 1

    for o in inputs_count.keys():
        inputs_count[o] = len(inputs_count[o])

    for w, attr in enumerate(basis_info):
        i, o = attr["in_irreps_position"], attr["out_irreps_position"]
        in_irrep, out_irrep = attr["in_irrep"], attr["out_irrep"]
        vars[w] = 1. / math.sqrt(inputs_count[o] * basis_count[(in_irrep, o)])

    return vars


cached_he_vars = {}

def generalized_he_init(tensor: torch.Tensor, basismanager: BasisManager, cache: bool = False):
    r"""
    
    Initialize the weights of a convolutional layer with a generalized He's weight initialization method.
    
    Because the computation of the variances can be expensive, to save time on consecutive runs of the same model,
    it is possible to cache the tensor containing the variance of each weight, for a specific ```basismanager```.
    This can be useful if a network contains multiple convolution layers of the same kind (same input and output types,
    same kernel size, etc.) or if one needs to train the same network from scratch multiple times (e.g. hyper-parameter
    search over learning rate).

    .. note ::
        The variance tensor is cached in memory and therefore is only available to the current process.
    
    Args:
        tensor (torch.Tensor): the tensor containing the weights
        basismanager (BasisManager): the basis expansion method
        cache (bool, optional): cache the variance tensor. By default, ```cache=False```

    """
    # Initialization
    
    assert tensor.shape == (basismanager.dimension(), )

    if cache and basismanager not in cached_he_vars:
        cached_he_vars[basismanager] = _generalized_he_init_variances(basismanager)

    if cache:
        vars = cached_he_vars[basismanager]
    else:
        vars = _generalized_he_init_variances(basismanager)

    tensor[:] = vars.to(device=tensor.device) * torch.randn_like(tensor)


def deltaorthonormal_init(tensor: torch.Tensor, basismanager: BasisManager):
    r"""
    
    Initialize the weights of a convolutional layer with *delta-orthogonal* initialization.

    If no 'radius' attribute is present in the `basismanager.get_basis_info()`, it is assumed the parameters
    parametrize a :class:`~escnn.nn.Linear` layer.
    
    Args:
        tensor (torch.Tensor): the tensor containing the weights
        basismanager (BasisManager): the basis expansion method

    """
    # Initialization

    assert tensor.shape == (basismanager.dimension(), )
    
    tensor.fill_(0.)
    
    counts = defaultdict(lambda: defaultdict(lambda: []))
    
    for p, attr in enumerate(basismanager.get_basis_info()):
        i = attr["in_irrep"]
        o = attr["out_irrep"]
        ip = attr["in_irreps_position"]
        op = attr["out_irreps_position"]
        if 'radius' in attr:
            r = attr["radius"]
        else:
            # assume this is the Linear layer
            r = 0.
        
        if i == o and math.isclose(r, 0.):
            counts[i][(ip, op)].append(p)
    
    def same_content(l):
        l = list(l)
        return all(ll == l[0] for ll in l)
    
    for irrep, count in counts.items():
        assert same_content([len(x) for x in count.values()]), [len(x) for x in count.values()]
        in_c = defaultdict(int)
        out_c = defaultdict(int)
        for ip, op in count.keys():
            in_c[ip] += 1
            out_c[op] += 1
        assert same_content(in_c.values()), count.keys()
        assert same_content(out_c.values()), count.keys()
        
        assert list(in_c.values())[0] == len(out_c.keys())
        assert list(out_c.values())[0] == len(in_c.keys())
        
        s = len(list(count.values())[0])
        i = len(in_c.keys())
        o = len(out_c.keys())
        
        # assert i <= o, (i, o, s, irrep, self._input_size, self._output_size)
        # if i > o:
        #     print("Warning: using delta orthogonal initialization to map to a larger number of channels")
        
        if max(o, i) > 1:
            W = stats.ortho_group.rvs(max(i, o))[:o, :i]
            # W = np.eye(o, i)
        else:
            W = 2 * torch.randint(0, 1, size=(1, 1)) - 1
            # W = np.array([[1]])
        
        w = torch.randn((o, s))
        # w = torch.ones((o, s))
        w *= 5.
        w /= (w ** 2).sum(dim=1, keepdim=True).sqrt()
        
        for i, ip in enumerate(in_c.keys()):
            for o, op in enumerate(out_c.keys()):
                for p, pp in enumerate(count[(ip, op)]):
                    tensor[pp] = w[o, p] * W[o, i]





from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from escnn.nn.modules.equivariant_module import EquivariantModule
from escnn.nn.modules.utils import indexes_from_labels

import torch

from typing import List, Tuple, Any
from collections import defaultdict
import numpy as np

__all__ = ["NormPool"]


class NormPool(EquivariantModule):
    
    def __init__(self, in_type: FieldType, **kwargs):
        r"""
        
        Module that implements Norm Pooling.
        For each input field, an output one is built by taking the norm of that field; as a result, the output
        field transforms according to a trivial representation.
        
        Args:
            in_type (FieldType): the input field type
            
        """
        assert isinstance(in_type.gspace, GSpace)
        
        super(NormPool, self).__init__()

        self.space = in_type.gspace
        self.in_type = in_type
        
        # build the output representation substituting each input field with a trivial representation
        self.out_type = FieldType(self.space, [self.space.trivial_repr] * len(in_type))

        # indices of the channels corresponding to fields belonging to each group in the input representation
        _in_indices = defaultdict(lambda: [])
        # indices of the channels corresponding to fields belonging to each group in the output representation
        _out_indices = defaultdict(lambda: [])

        # whether each group of fields is contiguous or not
        self._contiguous = {}

        # group fields by their size and
        #   - check if fields of the same size are contiguous
        #   - retrieve the indices of the fields
        indeces = indexes_from_labels(in_type, [r.size for r in in_type.representations])

        for s, (contiguous, fields, idxs) in indeces.items():
            self._contiguous[s] = contiguous
            if contiguous:
                # for contiguous fields, only the first and last indices are kept
                _in_indices[s] = torch.LongTensor([min(idxs), max(idxs)+1])
                _out_indices[s] = torch.LongTensor([min(fields), max(fields)+1])
            else:
                # otherwise, transform the list of indices into a tensor
                _in_indices[s] = torch.LongTensor(idxs)
                _out_indices[s] = torch.LongTensor(fields)
        
            # register the indices tensors as parameters of this module
            self.register_buffer('in_indices_{}'.format(s), _in_indices[s])
            self.register_buffer('out_indices_{}'.format(s), _out_indices[s])

    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""
        
        Apply the Norm Pooling to the input feature map.
        
        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map
            
        """
        
        assert input.type == self.in_type

        coords = input.coords
        input = input.tensor
        b, c = input.shape[:2]
        spatial_shape = input.shape[2:]

        output = torch.empty(self.evaluate_output_shape(input.shape), device=input.device, dtype=torch.float)
        
        for s, contiguous in self._contiguous.items():
            
            in_indices = getattr(self, f"in_indices_{s}")
            out_indices = getattr(self, f"out_indices_{s}")
            
            if contiguous:
                fm = input[:, in_indices[0]:in_indices[1], ...]
            else:
                fm = input[:, in_indices, ...]
                
            # split the channel dimension in 2 dimensions, separating fields
            fm = fm.view(b, -1, s, *spatial_shape)
            
            if contiguous:
                output[:, out_indices[0]:out_indices[1], ...] = fm.norm(dim=2)
            else:
                output[:, out_indices, ...] = fm.norm(dim=2)
        
        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, coords)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:
        assert len(input_shape) > 1
        assert input_shape[1] == self.in_type.size
    
        b, c = input_shape[:2]
        spatial_shape = input_shape[2:]

        return (b, self.out_type.size, *spatial_shape)

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
    
        c = self.in_type.size
    
        x = torch.randn(3, c, *[10]*self.in_type.gspace.dimensionality)
    
        x = GeometricTensor(x, self.in_type)
    
        errors = []
    
        for el in self.space.testing_elements:
            out1 = self(x).transform_fibers(el)
            out2 = self(x.transform_fibers(el))
        
            errs = (out1.tensor - out2.tensor).detach().numpy()
            errs = np.abs(errs).reshape(-1)
            print(el, errs.max(), errs.mean(), errs.var())
        
            assert torch.allclose(out1.tensor, out2.tensor, atol=atol, rtol=rtol), \
                'The error found during equivariance check with element "{}" is too high: max = {}, mean = {} var ={}' \
                    .format(el, errs.max(), errs.mean(), errs.var())
        
            errors.append((el, errs.mean()))
    
        return errors






from __future__ import annotations

from typing import Tuple, Callable, Iterable, List, Any, Dict

import escnn.group

from escnn.group import Group, GroupElement, IrreducibleRepresentation
from escnn.group.irrep import restrict_irrep

import numpy as np
import itertools
import re

__all__ = [
    "DirectProductGroup",
    'direct_product'
]


param_split_regex = re.compile(r"\[(.+)\ \|\ (.+)\]")


def _split_param(param: str) -> Tuple[str, str]:
    if param is None:
        return None, None
    
    params = param_split_regex.match(param).groups()
    assert len(params) == 2
    return params


class DirectProductGroup(Group):

    def __init__(self, G1: str, G2: str, name: str = None, **groups_keys):
        r"""
        
        Class defining the direct product of two groups.
        
        .. warning::
            This class should not be directly instantiated to ensure caching is performed correclty.
            You should instead use the function :func:`~escnn.group.direct_product`.
        
        .. warning::
            This class does not support all possible subgroups of the direct product!
            For :math:`G = G_1 \times G_2`, only subgroups of the form :math:`H = H_1 \times H_2` with :math:`H_1 < G_1`
            and :math:`H_2 < G_2` are supported.
            
            A subgroup `id` is a tuple containing a pair `(id1, id2)`, where `id1` identifies a subgroup of :math:`G_1`
            while `id2` identifies a subgroup of :math:`G_2`.
            See the documentation of the two groups used for more details about their subgroup structures.
            
            In case either :math:`H_1` or :math:`H_2` is the trivial subgroup (containing only the identity),
            the subgroup returned is just an instance of the other one and not of
            :class:`~escnn.group.DirectProductGroup`.
            
        
        Args:
            G1 (Group): first group
            G2 (Group): second group
            name (str, optional): name assigned to the resulting group
            groups_keys: additional keywords argument used for identifying the groups and perform caching
            
        Attributes:
            ~.G1 (Group): the first group
            ~.G2 (Group): the second group

        """
        
        g1_keys = {
            k[3:]: v
            for k, v in groups_keys.items()
            if k[:3] == 'G1_'
        }
        g2_keys = {
            k[3:]: v
            for k, v in groups_keys.items()
            if k[:3] == 'G2_'
        }
        
        self._G1 = escnn.group.groups_dict[G1]._generator(**g1_keys)
        self._G2 = escnn.group.groups_dict[G2]._generator(**g2_keys)
        
        continuous = self.G1.continuous or self.G2.continuous
        abelian = self.G1.abelian and self.G2.abelian
        
        self._defaulf_name = name is None
        if name is None:
            name = self.G1.name + ' X ' + self.G2.name
        
        assert isinstance(name, str)
        
        super(DirectProductGroup, self).__init__(name, continuous, abelian)
        
        self._build_representations()

    @property
    def G1(self) -> Group:
        return self._G1
    
    @property
    def G2(self) -> Group:
        return self._G2

    @property
    def PARAM(self) -> str:
        return f"[{self.G1.PARAM} | {self.G2.PARAM}]"

    @property
    def PARAMETRIZATIONS(self) -> List[str]:
        params = []
        for p1 in self.G1.PARAMETRIZATIONS:
            for p2 in self.G2.PARAMETRIZATIONS:
                params.append(
                    f"[{p1} | {p2}]"
                )
        return params

    @property
    def identity(self) -> GroupElement:
        return self.pair_elements(self.G1.identity, self.G2.identity)

    @property
    def elements(self) -> List[GroupElement]:
        e1 = self.G1.elements
        e2 = self.G2.elements
        if e1 is None or e2 is None:
            return None
        else:
            elements = []
            for g1 in e1:
                for g2 in e2:
                    elements.append(
                        self.pair_elements(g1, g2)
                    )
            return elements

    @property
    def subgroup_trivial_id(self):
        r"""
            The subgroup id associated with the trivial subgroup containing only the identity element :math:`{e}`.
            The id can be used in the method :meth:`~escnn.group.Group.subgroup` to generate the subgroup.
        """
        return (self.G1.subgroup_trivial_id, self.G2.subgroup_trivial_id)
    
    @property
    def subgroup_self_id(self):
        r"""
            The subgroup id associated with the group itself.
            The id can be used in the method :meth:`~escnn.group.Group.subgroup` to generate the subgroup.
        """
        return (self.G1.subgroup_self_id, self.G2.subgroup_self_id)

    @property
    def subgroup1_id(self):
        r"""
            The subgroup id associated with :math:`G1`.
        """
        return (self.G1.subgroup_self_id, self.G2.subgroup_trivial_id)

    @property
    def subgroup2_id(self):
        r"""
            The subgroup id associated with :math:`G2`.
        """
        return (self.G1.subgroup_trivial_id, self.G2.subgroup_self_id)

    @property
    def _keys(self) -> Dict[str, Any]:
        keys = dict()
        keys['G1'] = self.G1.__class__.__name__
        keys['G2'] = self.G2.__class__.__name__
        if not self._defaulf_name:
            keys['name'] = self.name
            
        keys.update({
            'G1_' + k: v
            for k, v in self.G1._keys.items()
        })
        keys.update({
            'G2_' + k: v
            for k, v in self.G2._keys.items()
        })
        return keys

    @property
    def generators(self) -> List[GroupElement]:
        return [
            self.inclusion1(g) for g in self.G1.generators
        ] + [
            self.inclusion2(g) for g in self.G2.generators
        ]
    
    def inclusion1(self, g1: GroupElement):
        return self.pair_elements(g1, self.G2.identity)
    
    def inclusion2(self, g2: GroupElement):
        return self.pair_elements(self.G1.identity, g2)

    def pair_elements(self, g1: GroupElement, g2: GroupElement):
        assert g1.group == self.G1
        assert g2.group == self.G2
        return self.element((
            g1.to(self.G1.PARAM),
            g2.to(self.G2.PARAM),
            self.PARAM
        ))

    def split_element(self, g: GroupElement) -> Tuple[GroupElement, GroupElement]:
        assert g.group == self
        g1, g2 = g.to(self.PARAM)
        return (
            self.G1.element(g1),
            self.G2.element(g2)
        )

    ###########################################################################
    # METHODS DEFINING THE GROUP LAW AND THE OPERATIONS ON THE GROUP'S ELEMENTS
    ###########################################################################

    def _combine(self, e1, e2,
                 param: str = None,
                 param1: str = None,
                 param2: str = None
                 ):
        r"""

        Method that returns the combination of two group elements according to the *group law*.
        
        Args:
            e1: an element of the group
            e2: another element of the group
    
        Returns:
            the group element :math:`e_1 \cdot e_2`
            
        """
        if param is None:
            param = self.PARAM
            
        param_1, param_2 = _split_param(param)
        param1_1, param1_2 = _split_param(param1)
        param2_1, param2_2 = _split_param(param2)
        
        return (
            self.G1._combine(e1[0], e2[0], param=param_1, param1=param1_1, param2=param2_1),
            self.G2._combine(e1[1], e2[1], param=param_2, param1=param1_2, param2=param2_2),
        )

    def _inverse(self, element, param: str = None):
        r"""
        Method that returns the inverse in the group of the element given as input

        Args:
            element: an element of the group

        Returns:
            its inverse
        """
        if param is None:
            param = self.PARAM

        param_1, param_2 = _split_param(param)

        return (
            self.G1._inverse(element[0], param=param_1),
            self.G2._inverse(element[1], param=param_2)
        )

    def _is_element(self,
                    element,
                    param: str = None,
                    verbose: bool = False
                    ) -> bool:
        r"""
        Check whether the input is an element of this group or not.

        Args:
            element: input object to test

        Returns:
            if the input is an element of the group

        """
        if param is None:
            param = self.PARAM

        param_1, param_2 = _split_param(param)

        return (
                self.G1._is_element(element[0], param=param_1, verbose=verbose) and
                self.G2._is_element(element[1], param=param_2, verbose=verbose)
        )

    def _equal(self, e1, e2,
               param: str = None,
               param1: str = None,
               param2: str = None
               ) -> bool:
        r"""
        Method that checks whether the two inputs are the same element of the group.

        This is especially useful for continuous groups with periodicity; see for instance
        :meth:`escnn.group.SO2.equal`.

        Args:
            e1: an element of the group
            e2: another element of the group

        Returns:
            if they are equal

        """
        if param is None:
            param = self.PARAM

        param_1, param_2 = _split_param(param)
        param1_1, param1_2 = _split_param(param1)
        param2_1, param2_2 = _split_param(param2)

        return (
            self.G1._equal(e1[0], e2[0], param=param_1, param1=param1_1, param2=param2_1) and
            self.G2._equal(e1[1], e2[1], param=param_2, param1=param1_2, param2=param2_2)
        )

    def _change_param(self, element, p_from: str, p_to: str):
        p_from_1, p_from_2 = _split_param(p_from)
        p_to_1, p_to_2 = _split_param(p_to)
    
        return (
                self.G1._change_param(element[0], p_from=p_from_1, p_to=p_to_1),
                self.G2._change_param(element[1], p_from=p_from_2, p_to=p_to_2)
        )

    def _hash_element(self, element, param: str = None):
        r"""
        Method that returns a unique hash for a group element given in input

        Args:
            element: an element of the group

        Returns:
            a unique hash
        """
        if param is None:
            param = self.PARAM

        param_1, param_2 = _split_param(param)

        return (
            self.G1._hash_element(element[0], param=param_1) +
            self.G2._hash_element(element[1], param=param_2)
        )

    def _repr_element(self, element, param: str = None):
        r"""
        Method that returns a representative string for a group element given in input

        Args:
            element: an element of the group

        Returns:
            a unique hash
        """
        if param is None:
            param = self.PARAM

        param_1, param_2 = _split_param(param)

        return (
                self.G1._repr_element(element[0], param=param_1) +
                ' ; ' +
                self.G2._repr_element(element[1], param=param_2)
        )

    ###########################################################################

    def __eq__(self, other):
        if not isinstance(other, DirectProductGroup):
            return False
        else:
            return self.G1 == other.G1 and self.G2 == other.G2

    def sample(self):
        return self.element((
            self.G1.sample().to(self.G1.PARAM),
            self.G2.sample().to(self.G2.PARAM)
        ))

    def grid(self, type: str = None, N: int = None, **kwargs):
        r"""

            .. todo ::
                Write documentation

            if `type = "rand"`, generates `N` random samples otherwise split the keywords between the two groups
            (check if they start with `"G1_"` or `"G2_"`) and generate the direct product of the two sets


        """

        if type == 'rand':
            assert N is not None and N >= 0

            grid1 = self.G1.grid(type='rand', N=N)
            grid2 = self.G2.grid(type='rand', N=N)

            return [
                self.element((
                    g1.to(self.G1.PARAM),
                    g2.to(self.G2.PARAM)
                ))
                for g1, g2 in zip(grid1, grid2)
            ]
        else:

            kwargs1 = {
                k[(len('G1_')):] : v
                for k, v in kwargs.items()
                if k.startswith('G1_')
            }
            kwargs2 = {
                k[(len('G2_')):] : v
                for k, v in kwargs.items()
                if k.startswith('G2_')
            }

            grid1 = self.G1.grid(**kwargs1)
            grid2 = self.G2.grid(**kwargs2)

            return [
                self.element((
                    g1.to(self.G1.PARAM),
                    g2.to(self.G2.PARAM)
                ))
                for g1 in grid1
                for g2 in grid2
            ]

    def _process_subgroup_id(self, id):
        r'''
        
        .warning::
            This class does not support all possible subgroups of the direct product!
            For :math:`G = G_1 \times G_2`, only subgroups of the form :math:`H = H_1 \times H_2` with :math:`H_1 < G_1`
            and :math:`H_2 < G_2` are supported.

        '''
        assert isinstance(id, tuple) and len(id) == 2
        id = (
            self.G1._process_subgroup_id(id[0]),
            self.G2._process_subgroup_id(id[1])
        )
        return id

    def _subgroup(self, id) -> Tuple[
        escnn.group.Group,
        Callable[[escnn.group.GroupElement], escnn.group.GroupElement],
        Callable[[escnn.group.GroupElement], escnn.group.GroupElement]
    ]:
        r"""
        Restrict the current group to the subgroup identified by the input ``id``.
        
        .warning::
            This class does not support all possible subgroups of the direct product!
            For :math:`G = G_1 \times G_2`, only subgroups of the form :math:`H = H_1 \times H_2` with :math:`H_1 < G_1`
            and :math:`H_2 < G_2` are supported.

        Args:
            id: the identifier of the subgroup

        Returns:
            a tuple containing

                - the subgroup,

                - a function which maps an element of the subgroup to its inclusion in the original group and

                -a function which maps an element of the original group to the corresponding element in the subgroup (returns None if the element is not contained in the subgroup)

        """
        
        H1, parent1, child1 = self.G1.subgroup(id[0])
        H2, parent2, child2 = self.G2.subgroup(id[1])
        
        if H1.order() == 1:
            return H2, inclusion_right(self, H2, parent2), restrict_right(self, H2, child2)
        elif H2.order() == 1:
            return H1, inclusion_left(self, H1, parent1), restrict_left(self, H1, child1)
        else:
            H = direct_product(H1, H2)
            return H, inclusion(self, H, parent1, parent2), restrict(self, H, child2, child2)

    def _combine_subgroups(self, sg_id1, sg_id2):
        assert isinstance(sg_id1, tuple) and len(sg_id1) == 2
        assert isinstance(sg_id2, tuple) and len(sg_id2) == 2

        id = (
            self.G1._combine_subgroups(sg_id1[0], sg_id2[0]),
            self.G2._combine_subgroups(sg_id1[1], sg_id2[1]),
        )
        return id

    @property
    def trivial_representation(self) -> escnn.group.IrreducibleRepresentation:
        r"""
        Builds the trivial representation of the group.
        The trivial representation is a 1-dimensional representation which maps any element to 1,
        i.e. :math:`\forall g \in G,\ \rho(g) = 1`.
        
        Returns:
            the trivial representation of the group

        """
        return self.irrep(
            self.G1.trivial_representation.id,
            self.G2.trivial_representation.id
        )

    def irrep(self, id1: Tuple, id2: Tuple, i: int = 0) -> escnn.group.IrreducibleRepresentation:
        r"""

        Builds the irreducible representation (:class:`~escnn.group.IrreducibleRepresentation`) of the group which is
        specified by the input arguments.

        .. seealso ::

            Check the documentation of the specific group subclass used for more information on the valid ``id`` values.

        Args:
            *id: parameters identifying the specific irrep.

        Returns:
            the irrep built

        """
        
        psi1 = self.G1.irrep(*id1)
        psi2 = self.G2.irrep(*id2)
        
        if psi1.type == 'C' and psi2.type == 'C':
            assert i in [0, 1]
        else:
            assert i == 0

        name = f"irrep_[{id1},{id2}]({i})"
        id = (id1, id2, i)
        if id not in self._irreps:
    
            if psi1.type == 'R' or psi2.type == 'R':
                assert i == 0
                type = psi2.type if psi1.type == 'R' else psi1.type
                size = psi1.size * psi2.size
                irrep, character = tensor_product_irrep(self, psi1, psi2)

            elif psi1.type == 'C' or psi2.type == 'C':
                assert i in [0, 1]
                type = 'C'
                size = psi1.size * psi2.size // 2
                irrep, character = tensor_product_irrep_complex(self, psi1, psi2, i)
                
            else:
                assert i == 0
                type = 'H'
                size = psi1.size * psi2.size // 2
                irrep, character = tensor_product_irrep_complex(self, psi1, psi2, 0)

            supported_nonlinearities = ['norm', 'gated']
            self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, size, type,
                                                         supported_nonlinearities=supported_nonlinearities,
                                                         character=character,
                                                         id1=id1,
                                                         id2=id2,
                                                         i=i)
        
        return self._irreps[id]

    def _restrict_irrep(self, irrep: Tuple, id) -> Tuple[np.matrix, List[Tuple]]:

        # compute the restriction numerically
        # it should be possible to derive the decomposition analitically for the special case in which the subgroup is
        # itself the direct product of two subgroups of the two groups composing this group.
        # While in the complex field it is rather trivial, in the real one it is much more complicated and requires
        # us to assume that the irrep restriction implemented by both subgroups resemble the "realification" of the
        # irrep decomposition in the complex field.
        # it is much easier to solve this numerically

        irr = self.irrep(*irrep)

        change_of_basis, irreps = restrict_irrep(irr, id)
        return change_of_basis, irreps

    def _build_representations(self):
        r"""
        Build the irreps for this group

        """
    
        # Build all the Irreducible Representations
        for psi1 in self.G1.irreps():
            for psi2 in self.G2.irreps():
                self.irrep(psi1.id, psi2.id, 0)
                if psi1.type == 'C' and psi2.type == 'C':
                    self.irrep(psi1.id, psi2.id, 1)

        # add all the irreps to the set of representations already built for this group
        self.representations.update(**{irr.name: irr for irr in self.irreps()})

    def testing_elements(self) -> Iterable[GroupElement]:
        r"""
        A finite number of group elements to use for testing.
        """
        for g1, g2 in itertools.product(self.G1.testing_elements(), self.G2.testing_elements()):
            yield self.pair_elements(g1, g2)

    def _decode_subgroup_id_pickleable(self, id: Tuple) -> Tuple:
        if isinstance(id, tuple):
            if id[0] == 'G':
                id = self.element(id[1], id[2])
            elif id[0] == 'G1':
                id = self.G1.element(id[1], id[2])
            elif id[0] == 'G2':
                id = self.G2.element(id[1], id[2])
            else:
                id = list(id)
                for i in range(len(id)):
                    id[i] = self._decode_subgroup_id_pickleable(id[i])
                id = tuple(id)

        return id

    def _encode_subgroup_id_pickleable(self, id: Tuple) -> Tuple:

        if isinstance(id, GroupElement):
            G = id.group
            if G == self:
                id = 'G', id.value, id.param
            elif G == self.G1:
                id = 'G1', id.value, id.param
            elif G == self.G2:
                id = 'G2', id.value, id.param
            else:
                raise ValueError
        elif isinstance(id, tuple):
            id = list(id)
            for i in range(len(id)):
                id[i] = self._encode_subgroup_id_pickleable(id[i])
            id = tuple(id)
        return id

    _cached_group_instance = dict()

    @classmethod
    def _generator(cls, G1: str, G2: str, **group_keys) -> 'DirectProductGroup':
        
        key = {
            'G1': G1,
            'G2': G2,
        }
        key.update(**group_keys)
        
        key = tuple(sorted(key.items()))

        if key not in cls._cached_group_instance:
            cls._cached_group_instance[key] = DirectProductGroup(G1, G2, **group_keys)
            
        cls._cached_group_instance[key]._build_representations()

        return cls._cached_group_instance[key]


def direct_product(G1: Group, G2: Group, name: str = None):
    r'''
    
    Generates the direct product of the two input groups `G1` and `G2`.
    
    Args:
        G1 (Group): first group
        G2 (Group): second group
        name (str, optional): name assigned to the resulting group

    Returns:
        an instance of :class:`~escnn.group.DirectProductGroup`
 
    '''
    
    group_keys = dict()
    group_keys.update(**{
        'G1_' + k: v
        for k, v in G1._keys.items()
    })
    group_keys.update(**{
        'G2_' + k: v
        for k, v in G2._keys.items()
    })
    return DirectProductGroup._generator(
        G1.__class__.__name__,
        G2.__class__.__name__,
        name=name,
        **group_keys
    )


#############################################
# SUBGROUPS MAPS
#############################################

# Restrict G1 X G2 to a general H1 X H2
# with Hi < Gi

def restrict(G: DirectProductGroup, H: DirectProductGroup, restrict1, restrict2):
    
    def _map(e: GroupElement, G=G, H=H, restrict1=restrict1, restrict2=restrict2):
        assert e.group == G
        
        e1, e2 = G.split_element(e)
        
        _e1 = restrict1(e1)
        _e2 = restrict2(e2)
        
        if _e1 is None or _e2 is None:
            return None

        try:
            return H.element(
                (_e1.value, _e2.value),
                param=f'[{_e1.param} | {_e2.param}]'
            )
        except ValueError:
            return None
    
    return _map


def inclusion(G: DirectProductGroup, H: DirectProductGroup, inclusion1, inclusion2):
    
    def _map(e: GroupElement, G=G, H=H, inclusion1=inclusion1, inclusion2=inclusion2):
        assert e.group == H
        
        e1, e2 = H.split_element(e)
        
        _e1 = inclusion1(e1)
        _e2 = inclusion2(e2)
        
        return G.element(
            (_e1.value, _e2.value),
            param=f'[{_e1.param} | {_e2.param}]'
        )
    
    return _map


# Restrict G1 X G2 to H X {e}  = H
# with H < G1


def restrict_left(G: DirectProductGroup, H: Group, restrict1):
    def _map(e: GroupElement, G=G, H=H, restrict1=restrict1):
        assert e.group == G
        
        e1, e2 = G.split_element(e)
        
        if e2 != G.G2.identity:
            return None
        
        return restrict1(e1)
    
    return _map


def inclusion_left(G: DirectProductGroup, H: Group, inclusion1):
    def _map(e: GroupElement, G=G, H=H, inclusion1=inclusion1):
        assert e.group == H
        
        _e1 = inclusion1(e)
        
        return G.element((
                _e1.value,
                G.G2.identity.value
            ),
            param=f"[{_e1.param} | {G.G2.identity.param}]"
        )

    return _map


# Restrict G1 X G2 to {e} X H  = H
# with H < G2

def restrict_right(G: DirectProductGroup, H: Group, restrict2):
    def _map(e: GroupElement, G=G, H=H, restrict2=restrict2):
        assert e.group == G
        
        e1, e2 = G.split_element(e)
        
        if e1 != G.G1.identity:
            return None
        
        return restrict2(e2)
    
    return _map


def inclusion_right(G: DirectProductGroup, H: Group, inclusion2):
    def _map(e: GroupElement, G=G, H=H, inclusion2=inclusion2):
        assert e.group == H
        
        _e2 = inclusion2(e)

        return G.element((
                G.G1.identity.value,
                _e2.value
            ),
            param=f"[{G.G1.identity.param} | {_e2.param}]"
        )

    return _map


########################################################################################################################
# Generate Irreps
########################################################################################################################

def tensor_product_irrep(G: DirectProductGroup, psi1: IrreducibleRepresentation, psi2: IrreducibleRepresentation):
    
    # for a direct product group G = A x B, if
    # alpha_i is an irrep of A
    # beta_j is an irrep of B
    # and at least one of alpha_i and beta_j is a real-type irrep
    # G has an irrep of the form
    #       gamma_ij = alpha_i x beta_j
    # the type of gamma_ij is the type of the other irrep among alpha_i and beta_j
    
    # Therefore, gamma_ij can be simply computed as the tensor product of alpha_i and beta_j

    assert psi1.group == G.G1
    assert psi2.group == G.G2
    
    size = psi1.size * psi2.size
    size2_half = psi2.size // 2
    
    def tensor_product_left(g: GroupElement, G=G, psi1=psi1, psi2=psi2, size=size, size2_half=size2_half):
        assert g.group == G
        g1, g2 = G.split_element(g)
        psi_g1 = psi1(g1)
        psi_g2 = psi2(g2)

        real_2 = psi_g2[:size2_half, :size2_half]
        imag_2 = psi_g2[size2_half:, :size2_half]

        R = np.empty((size, size))
        real = np.kron(psi_g1, real_2)
        imag = np.kron(psi_g1, imag_2)

        R[:size // 2, :size // 2] = real
        R[size // 2:, size // 2:] = real
        R[size // 2:, :size // 2] = imag
        R[:size // 2:, size // 2:] = -imag

        return R

    def tensor_product_right(g: GroupElement, G=G, psi1=psi1, psi2=psi2):
        assert g.group == G
        g1, g2 = G.split_element(g)
        return np.kron(psi1(g1), psi2(g2))

    def character(g: GroupElement, G=G, psi1=psi1, psi2=psi2):
        assert g.group == G
        g1, g2 = G.split_element(g)
        return psi1.character(g1) * psi2.character(g2)

    if psi1.type == 'R' and psi2.type != 'R':
        return tensor_product_left, character
    elif psi2.type == 'R':
        return tensor_product_right, character
    else:
        raise ValueError()


def tensor_product_irrep_complex(G: DirectProductGroup, psi1: IrreducibleRepresentation, psi2: IrreducibleRepresentation, i: int):
    
    # assume a direct product group G = A x B, with
    # alpha_i is an irrep of A
    # beta_j is an irrep of B
    # neither alpha_i nor beta_j are real type-type irreps; then
    # alpha_i is isomorphic to Realification(a_i)
    # beta_j is isomorphic to Realification(b_i)
    # where a_i is a complex irrep of A, b_j is a complex irrep of B
    # and "Realification(X)" is a real matrix with the following block structure
    #  +---------+----------+
    #  | Real(X) | -Imag(X) |
    #  +---------+----------+
    #  | Imag(X) |  Real(X) |
    #  +---------+----------+

    # G has a real irrep of the form
    #       gamma_ij = Realification(a_i x b_j)
    # and one of the form
    #       delta_ij = Realification(a_i x ~b_j)
    # where ~ is the complex conjugate
    # In particular,
    #       gamma_ij + delta_ij = alpha_i  x beta_j
    # where = represent isomorphism
    # Additionally, if either alpha_i or beta_j have quaternionic-type
    # (or, equivalently, a_i = ~a_i or b_j = ~b_j)
    # then gamma_ij = delta_ij
    
    # Finally, if both alpha_i and beta_j are quaternionic type, then so is gamma_ij (or delta_ij)
    # Otherwise, gamma_ij (and delta_ij) has complex-type.
    
    # To compute gamma_ij (i=0) and delta_ij (i=1), note the following:
    # Real(a_i x b_j) = Real(a_i) x Real(b_j) - Imag(a_i) x Imag(b_j)
    # Imag(a_i x b_j) = Real(a_i) x Imag(b_j) + Real(a_i) x Imag(b_j)
    
    # We assume that both alpha_i and beta_j are expressed in a basis such that
    # alpha_i = Realification(a_i)
    # beta_j = Realification(b_i)
    # where = is an equality here
    # This allows us to extract the real and the imaginary parts of a_i and b_j from alpha_i and beta_j, to then compute
    # gamma_ij using the expressions above and the definition of Realification()
    # The computation of delta_ij is equivalent, up to a change of sign of Imag(b_j) in all expressions.

    assert psi1.group == G.G1
    assert psi2.group == G.G2

    size1 = psi1.size
    size2 = psi2.size
    size = psi1.size * psi2.size // 2
    
    assert i in [0, 1]
    
    def tensor_product(g: GroupElement, G=G, psi1=psi1, psi2=psi2, i=i, size=size, size1=size1, size2=size2):
        assert g.group == G
        
        g1, g2 = G.split_element(g)
        
        psi_g1 = psi1(g1)
        psi_g2 = psi2(g2)
        
        real_1 = psi_g1[:size1//2, :size1//2]
        imag_1 = psi_g1[size1//2:, :size1//2]
        
        real_2 = psi_g2[:size2//2, :size2//2]
        imag_2 = psi_g2[size2//2:, :size2//2]
        
        if i == 1:
            imag_2 *= -1
            
        R = np.empty((size, size))
        real = np.kron(real_1, real_2) - np.kron(imag_1, imag_2)
        imag = np.kron(real_1, imag_2) + np.kron(imag_1, real_2)
        
        R[:size//2, :size//2] = real
        R[size//2:, size//2:] = real
        R[size//2:, :size//2] = imag
        R[:size//2:, size//2:] = -imag
        
        return R

    def character(g: GroupElement, G=G, psi1=psi1, psi2=psi2, i=i):
        assert g.group == G
        g1, g2 = G.split_element(g)
        
        psi_g1 = psi1(g1)
        psi_g2 = psi2(g2)

        real_1 = psi_g1[:size1 // 2, :size1 // 2]
        imag_1 = psi_g1[size1 // 2:, :size1 // 2]

        real_2 = psi_g2[:size2 // 2, :size2 // 2]
        imag_2 = psi_g2[size2 // 2:, :size2 // 2]

        if i == 1:
            imag_2 *= -1

        return 2 * (
                np.trace(real_1) * np.trace(real_2)
              - np.trace(imag_1) * np.trace(imag_2)
        )

    return tensor_product, character








from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from ..equivariant_module import EquivariantModule

import torch.nn.functional as F
import torch
from typing import List, Tuple, Any

__all__ = ["PointwiseDropout"]


class PointwiseDropout(EquivariantModule):
    
    def __init__(self,
                 in_type: FieldType,
                 p: float = 0.5,
                 inplace: bool = False
                 ):
        r"""
        
        Applies dropout to individual *channels* independently.
        
        This class is just a wrapper for :func:`torch.nn.functional.dropout` in an :class:`~escnn.nn.EquivariantModule`.
        
        Only representations supporting pointwise non-linearities are accepted as input field type.
        
        Args:
            in_type (FieldType): the input field type
            p (float, optional): dropout probability
            inplace (bool, optional): can optionally do the operation in-place. Default: ``False``
            
        """
    
        assert isinstance(in_type.gspace, GSpace)
        if p < 0 or p > 1:
            raise ValueError("dropout probability has to be between 0 and 1, but got {}".format(p))

        super(PointwiseDropout, self).__init__()

        for r in in_type.representations:
            assert 'pointwise' in r.supported_nonlinearities, \
                'Error! Representation "{}" does not support "pointwise" non-linearity'.format(r.name)

        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = in_type

        self.p = p
        self.inplace = inplace
        
    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""
        
        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map
            
        """
        
        assert input.type == self.in_type
        
        output = F.dropout(input.tensor, self.p, self.training, self.inplace)
        
        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, input.coords)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:
        assert len(input_shape) > 1
        assert input_shape[1] == self.in_type.size
        return input_shape
    
    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
        # return super(InnerBatchNorm, self).check_equivariance(atol=atol, rtol=rtol)
        pass

    def export(self):
        r"""
        Export this module to a normal PyTorch :class:`torch.nn.Dropout` module and set to "eval" mode.

        """
    
        self.eval()
        
        return torch.nn.Dropout(self.p, self.inplace).eval()





from escnn.group import *

import torch
import numpy as np

from typing import Tuple


__all__ = ["HarmonicPolynomialR3Generator"]


class HarmonicPolynomialR3Generator(torch.nn.Module):

    def __init__(self, L: int):
        r"""
            Module which computes the *harmonic polynomials* in :math:`\R^3` up to order `L`.

            This equivariant module takes a set of 3-dimensional points transforming according to the
            :meth:`~escnn.group.O3.standard_representation` of :math:`O(3)` and outputs :math:`(L+1)^2` dimensional
            feature vectors transforming like spherical harmonics according to
            :meth:`~escnn.group.O3.bl_sphere_representation` with `L=L`.

            Harmonic polynomial are related to the spherical harmonics.
            In particular, when harmonic polynomials are evaluated on the unit sphere, they match the spherical
            harmonics.
            Check the
            `Wikipedia page <https://en.wikipedia.org/wiki/Spherical_harmonics#Harmonic_polynomial_representation>`_
            about them.

        """

        super(HarmonicPolynomialR3Generator, self).__init__()

        self.G: O3 = o3_group(L)

        self.L = L
        self.rho = self.G.bl_sphere_representation(L)
        self.d = self.rho.size

        if self.L > 0:
            self.register_buffer(f'cob_1', torch.tensor(self.G.standard_representation().change_of_basis_inv, dtype=torch.float))

        for l in range(2, self.L+1):
            rho_l = self.G.irrep((l-1)%2, l-1).tensor(self.G.irrep(1, 1))
            d = 2*l+1
            cob = rho_l.change_of_basis_inv[-d:, :]

            p = cob @ cob.T
            assert np.allclose(p, np.eye(2*l+1), atol=1e-5, rtol=1e-5), l

            # to ensure normalization

            # this guarantees the column is normalized, i.e. corresponds to the central column of the Wigner D matrix
            cob *= np.sqrt((2*l-1)/l)

            self.register_buffer(f'cob_{l}', torch.tensor(cob, dtype=torch.float))

    def forward(self, points: torch.Tensor):
        assert points.shape[-1] == 3, points.shape
        shape = points.shape[:-1]

        features_0 = torch.ones(*shape, 1, device=points.device, dtype=points.dtype)

        if self.L == 0:
            return features_0

        feature_1 = torch.einsum('...i,ji->...j', points, getattr(self, 'cob_1'))

        features = [
            features_0,
            feature_1,
        ]

        for l in range(2, self.L+1):
            cob = getattr(self, f'cob_{l}')
            f = torch.einsum('...i,...j->...ij', features[-1], feature_1).reshape(*shape, -1)
            assert f.shape[-1] == 3*(2*l-1), (f.shape, l)
            f = torch.einsum('...i,ji->...j', f, cob)
            assert f.shape[-1] == 2*l+1, (f.shape, l)

            features.append(f)

        features = torch.cat(features, dim=-1)

        return features

    def check_equivariance(self, atol: float = 1e-5, rtol: float = 1e-3):

        device = self.cob_1.device

        N = 40
        points = torch.randn(N, 3, device=device)
        # radii = torch.norm(points, dim=-1).view(-1, 1)
        # points = points / radii
        # points[radii.view(-1) < 1e-3, :] = 0.

        sh = self(points)
        for _ in range(10):
            g = self.G.sample()

            sh_rot = self(points @ torch.tensor(self.G.standard_representation()(g).T, dtype=torch.float, device=device))
            rot_sh = sh @ torch.tensor(self.rho(g).T, dtype=torch.float, device=device)
            assert torch.allclose(rot_sh, sh_rot, atol=atol, rtol=rtol), (rot_sh - sh_rot).abs().max().item()





from collections import defaultdict


from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from ..equivariant_module import EquivariantModule

import torch
from torch.nn import Parameter
from typing import List, Tuple, Any


__all__ = ["InducedNormBatchNorm"]


class InducedNormBatchNorm(EquivariantModule):
    
    def __init__(self,
                 in_type: FieldType,
                 eps: float = 1e-05,
                 momentum: float = 0.1,
                 affine: bool = True,
                 ):
        r"""

        Batch normalization for induced isometric representations.
        This module requires the input fields to be associated to an induced representation from an isometric
        (i.e. which preserves the norm) non-trivial representation which supports 'norm' non-linearities.
        
        The module assumes the mean of the vectors is always zero so no running mean is computed and no bias is added.
        This is guaranteed as long as the representations do not include a trivial representation.
        
        Indeed, if :math:`\rho` does not include a trivial representation, it holds:
        
        .. math ::
        
             \forall \bold{v} \in \mathbb{R}^n, \ \ \frac{1}{|G|} \sum_{g \in G} \rho(g) \bold{v} = \bold{0}

        Hence, only the standard deviation is normalized.
        The same standard deviation, however, is shared by all the sub-fields of the same induced field.
        
        The input representation of the fields is preserved by this operation.
        
        
        Only representations which do not contain the trivial representation are allowed.
        You can check if a representation contains the trivial representation using
        :meth:`~escnn.group.Representation.contains_trivial`.
        To check if a trivial irrep is present in a representation in a :class:`~escnn.nn.FieldType`, you can use::
        
            for r in field_type:
                if r.contains_trivial():
                    print(f"field type contains a trivial irrep")

        Args:
            in_type (FieldType): the input field type
            eps (float, optional): a value added to the denominator for numerical stability. Default: ``1e-5``
            momentum (float, optional): the value used for the ``running_mean`` and ``running_var`` computation.
                    Can be set to ``None`` for cumulative moving average (i.e. simple average). Default: ``0.1``
            affine (bool, optional): if ``True``, this module has learnable scale parameters. Default: ``True``

        """
    
        assert isinstance(in_type.gspace, GSpace)
        
        super(InducedNormBatchNorm, self).__init__()

        for r in in_type.representations:
            assert any(nl.startswith('induced_norm') for nl in r.supported_nonlinearities), \
                'Error! Representation "{}" does not support "induced_norm" non-linearity'.format(r.name)
            # Norm batch-normalization assumes the fields to have mean 0. This is true as long as it doesn't contain
            # the trivial representation
            for irr in r.irreps:
                assert not in_type.fibergroup._irreps[irr].is_trivial()

        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = in_type
        
        self.affine = affine
        
        # group fields by their size and
        #   - check if fields of the same size are contiguous
        #   - retrieve the indices of the fields

        # number of fields of each size
        self._nfields = defaultdict(int)
        
        # indices of the channales corresponding to fields belonging to each group
        _indices = defaultdict(lambda: [])
        
        # whether each group of fields is contiguous or not
        self._contiguous = {}
        
        position = 0
        last_id = None
        for i, r in enumerate(self.in_type.representations):
            subfield_size = None
            for nl in r.supported_nonlinearities:
                if nl.startswith('induced_norm'):
                    assert subfield_size is None, "Error! The representation supports multiple " \
                                                  "sub-fields of different sizes"
                    subfield_size = int(nl.split('_')[-1])
                    assert r.size % subfield_size == 0
    
            id = (r.size, subfield_size)
    
            if id != last_id:
                self._contiguous[id] = not id in self._contiguous
            last_id = id
            
            _indices[id] += list(range(position, position + r.size))
            self._nfields[id] += 1
            position += r.size
        
        for id, contiguous in self._contiguous.items():
            if contiguous:
                # for contiguous fields, only the first and last indices are kept
                _indices[id] = torch.LongTensor([min(_indices[id]), max(_indices[id])+1])
            else:
                # otherwise, transform the list of indices into a tensor
                _indices[id] = torch.LongTensor(_indices[id])
                
            # register the indices tensors as parameters of this module
            self.register_buffer(f'{id}_indices', _indices[id])
            
            running_var = torch.ones((1, self._nfields[id], 1, 1), dtype=torch.float)
            self.register_buffer(f'{id}_running_var', running_var)
            
            if self.affine:
                weight = Parameter(torch.ones((1, self._nfields[id], 1, 1)), requires_grad=True)
                self.register_parameter(f'{id}_weight', weight)
        
        _indices = dict(_indices)
        
        self._order = list(_indices.keys())
        
        self.register_buffer('num_batches_tracked', torch.tensor(0, dtype=torch.long))
        
        self.eps = eps
        self.momentum = momentum

        # self.weight = Parameter(torch.ones(len(in_type)), requires_grad=True)

    def reset_running_stats(self):
        for s in self._order:
            running_var = getattr(self, f"{s}_running_var")
            running_var.fill_(1)
        self.num_batches_tracked.zero_()

    def reset_parameters(self):
        self.reset_running_stats()
        for s in self._order:
            weight = getattr(self, f"{s}_weight")
            weight.data.fill_(1)
    
    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""
        Apply norm non-linearities to the input feature map
        
        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map
            
        """
        
        assert input.type == self.in_type

        exponential_average_factor = 0.0

        if self.training:
            self.num_batches_tracked += 1
            if self.momentum is None:  # use cumulative moving average
                exponential_average_factor = 1.0 / self.num_batches_tracked.item()
            else:  # use exponential moving average
                exponential_average_factor = self.momentum

        # compute the squares of the values of each channel
        # n = torch.mul(input.tensor, input.tensor)
        n = input.tensor.detach()**2

        coords = input.coords
        b, c = input.tensor.shape[:2]
        shape = input.tensor.shape[2:]

        output = input.tensor.clone()
        
        if self.training:
            
            # self.running_var *= 1 - exponential_average_factor
            
            next_var = 0
            # iterate through all field sizes
            for id in self._order:
                size, subfield_size = id
                n_subfields = int(size // subfield_size)
                
                indices = getattr(self, f"{id}_indices")
                running_var = getattr(self, f"{id}_running_var")
                
                # compute the norm squared of the fields
                
                if self._contiguous[id]:
                    # if the fields were contiguous, we can use slicing

                    # compute the norm of each field by summing the squares
                    norms = n[:, indices[0]:indices[1], ...] \
                        .view(b, -1, n_subfields, subfield_size, *shape) \
                        .sum(dim=3, keepdim=False) #.sqrt()
                else:
                    # otherwise we have to use indexing
            
                    # compute the norm of each field by summing the squares
                    norms = n[:, indices, ...] \
                        .view(b, -1,  n_subfields, subfield_size, *shape) \
                        .sum(dim=3, keepdim=False) #.sqrt()
                
                # Since the mean of the fields is 0, we can compute the variance as the mean of the norms squared
                # corrected with Bessel's correction
                norms = norms.transpose(0, 1).reshape(self._nfields[id], -1)
                correction = norms.shape[1]/(norms.shape[1]-1) if norms.shape[1] > 1 else 1
                vars = norms.mean(dim=1).view(1, -1, 1, 1) / subfield_size
                vars *= correction
                # vars = norms.transpose(0, 1).reshape(self._nfields[s], -1).var(dim=1)
        
                # self.running_var[next_var:next_var + self._nfields[s]] += exponential_average_factor * vars
                running_var *= 1 - exponential_average_factor
                running_var += exponential_average_factor * vars #.detach()

                next_var += self._nfields[id]

            # self.running_var = self.running_var.detach()
            
        next_var = 0
        
        # iterate through all field sizes
        for id in self._order:
            size, subfield_size = id
            n_subfields = int(size // subfield_size)
    
            indices = getattr(self, f"{id}_indices")
            
            # retrieve the running variances corresponding to the current fields
            vars = getattr(self, f"{id}_running_var")
            if self.affine:
                weight = getattr(self, f"{id}_weight")
            else:
                weight = 1.
            
            # compute the scalar multipliers needed
            multipliers = weight / (vars + self.eps).sqrt()
            
            # expand the multipliers tensor to all channels for each field
            multipliers = multipliers.expand(b, -1, n_subfields, subfield_size).reshape(b, -1, *[1]*len(shape))

            if self._contiguous[id]:
                # if the fields are contiguous, we can use slicing
                output[:, indices[0]:indices[1], ...] *= multipliers
            else:
                # otherwise we have to use indexing
                output[:, indices, ...] *= multipliers
            
            # shift the position on the running_var and weight tensors
            next_var += self._nfields[id]
        
        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, input.coords)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:

        assert len(input_shape) >= 2
        assert input_shape[1] == self.in_type.size

        b, c = input_shape[:2]
        spatial_shape = input_shape[2:]

        return (b, self.out_type.size, *spatial_shape)

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
        # return super(NormBatchNorm, self).check_equivariance(atol=atol, rtol=rtol)
        pass





from typing import Tuple, Union, List

from escnn.group import IrreducibleRepresentation, GroupElement, Group, Representation, directsum

import numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import connected_components


__all__ = ["HomSpace"]


class HomSpace:
    
    def __init__(self,
                 G: Group,
                 sgid: Tuple,
                 ):
        r"""
        Class defining an homogeneous space, i.e. the quotient space :math:`X \cong G / H` generated by a group
        :math:`G` and a subgroup :math:`H<G`, called the *stabilizer* subgroup.
        
        As a quotient space, the homogeneous space is defined as the set

        .. math::
            X \cong G / H = \{gH \ | g \in G \}
        
        where :math:`gH = \{gh | h \in H\}` is a *coset*.
        
        A classical example is given by the sphere :math:`S^2`, which can be interpreted as the quotient space
        :math:`S^2 \cong \SO3 / \SO2`, where :math:`\SO3` is the group of all 3D rotations and :math:`\SO2` here
        represents the subgroup of all planar rotations around the Z axis.
        
        This class is useful to generate bases for the space of functions or vector fields (Mackey functions) over the
        homogeneous space :math:`X\cong G / H`.
        
        Args:
            G (Group): the symmetry group of the space
            sgid (tuple): the id of the stabilizer subgroup
        """
        
        super(HomSpace, self).__init__()
        
        # Group:
        self.G = G
        
        self.H, self._inclusion, self._restriction = self.G.subgroup(sgid)

        # tuple:
        self.sgid = sgid

        # dict:
        self._representations = {}

        # dict:
        self._names_to_irreps = {}

        # dict:
        self._names_to_psi = {}

    def same_coset(self, g1: GroupElement, g2: GroupElement) -> bool:
        f'''
            Check if the input elements `g1` and `g2` belong to the same coset in :math:`G/H`, i.e. if
            :math:`\exists h : g_1 = g_2 h`.
        '''
        
        assert g1.group == self.G
        assert g2.group == self.G
        
        d = ~g1 @ g2
        
        return self._restriction(d) is not None

    def basis(self,
              g: GroupElement,
              rho: Union[IrreducibleRepresentation, Tuple],
              psi: Union[IrreducibleRepresentation, Tuple]
              ) -> np.ndarray:
        r"""
        
        Let `rho` be an irrep of :math:`G` and `psi` an irrep of :math:`H`.
        This method generates a basis for the subspace of `psi`-vector fields over :math:`X\cong G/H` which transforms
        under `rho` and samples its elements on the input :math:`g \in G`.

        .. note::
            Note that a `psi`-vector field :math:`f` is interpreted as a Mackey function, i.e. as a map
            :math:`f: G \to \R^{\text{dim}_\psi}`, rather than :math:`f: X \to \R^{\text{dim}_\psi}`.
            Indeed, this function takes an element :math:`g\in G` in input.
            This function can be composed with a choice of *section* :math:`\gamma: X \to G` to obtain a vector field
            over `X`.
        
        Let :math:`m` be the multiplicity of :math:`\rho`, i.e. the number of subspaces which transform according to
        `rho` and :math:`\text{dim}_\rho` the dimensionality of the :math:`G`-irrep `rho`,
        Then, the space is :math:`\text{dim}_\rho \times m` dimensional and this method generates
        :math:`\text{dim}_\rho \times m` basis functions over :math:`G`.
        
        However, this method returns an array of shape :math:`\text{dim}_\rho \times m \times \text{dim}_\psi`, where
        :math:`\text{dim}_\psi` is the dimensionality of the :math:`H`-irrep `psi`.
        Any slice along the last dimension of size :math:`\text{dim}_\psi`, returns a valid
        :math:`\text{dim}_\rho \times m` dimensional basis.
        
        The redundant elements along the last dimension can be used to express the :math:`H`-equivariance property of
        the Mackey functions in a convenient way.
        A Mackey function :math:`f` satisfies the following constraint
        
        .. math::
        
            f(g h) = \psi(h^{-1}) f(g)
        
        In the basis generated by this method, this action translates in the following property.
        A left multiplication by :math:`\rho(ghg^{-1})` along the first dimension is equivalent to a right multiplication by
        :math:`\psi(h)` along the last dimension.
        In other words::
        
            B = self.basis(g, rho, psi)
            Bh = np.einsum('ijp,pq->ijq', B, psi(h))
            hB = np.einsum('oi,ijp->ojp', rho(g) @ rho.restrict(self.sgid)(h) @ rho(g).T, B)
            assert np.allclose(Bh, hB)
        
        Args:
            g (GroupElement): the group element where to sample the elements of the basis
            rho (IrreducibleRepresentation): an irrep of `G` (or its id)
            psi (IrreducibleRepresentation): an irrep of `H` (or its id)

        Returns:
            an array of shape :math:`\text{dim}_\rho \times m \times \text{dim}_\psi` representing the basis elements
            sampled on `g`

        """
        
        assert g.group == self.G
        
        if isinstance(rho, tuple):
            rho = self.G.irrep(*rho)
            
        if isinstance(psi, tuple):
            psi = self.H.irrep(*psi)
        
        assert isinstance(rho, IrreducibleRepresentation)
        assert isinstance(psi, IrreducibleRepresentation)
        
        assert rho.group == self.G
        assert psi.group == self.H
        
        # (rho.size, multiplicity of rho in Ind psi, psi.size)
        # B[:, j, :] is an intertwiner between f(e) \in V_psi and the j-th occurrence of rho in Ind psi
        #
        # B_0(g) = rho(g) @ B[:, :, 0]
        # contains the basis for f \in Ind psi interpreted as a scalar function f: G \to R
        # (as a subrepresentation of the regular repr of G)
        # i.e. it contains a basis for f(g)_0
        #
        # The full tensor B(g) = rho(g) @ B
        # is a basis for f interpreted as a Mackey function f: G \to V_psi
        B = self._dirac_kernel_ft(rho.id, psi.id)
        
        # rho(g) @ B
        return np.einsum('oi, ijp->ojp', rho(g), B)
        
    def _dirac_kernel_ft(self, rho: Tuple, psi: Tuple, eps: float = 1e-9) -> np.ndarray:
        
        # TODO: this can be cached
        
        rho = self.G.irrep(*rho)
        psi = self.H.irrep(*psi)
        
        rho_H = rho.restrict(self.sgid)

        m_psi = 0
        for irrep in rho_H.irreps:
            if self.H.irrep(*irrep) == psi:
                m_psi += 1
                
        basis = np.zeros((rho.size, m_psi * psi.sum_of_squares_constituents, psi.size))
        
        # pick the arbitrary basis element e_i (i=0) for V_\psi
        i = 0
        
        p = 0
        j = 0
        
        column_mask = np.zeros(rho.size, dtype=np.bool)
        
        for irrep in rho_H.irreps:
            irrep = self.H.irrep(*irrep)
            
            if irrep == psi:
                w_i = (psi.endomorphism_basis()[:, i, :] **2).sum(axis=0)
                nonnull_mask = w_i > eps
                
                assert nonnull_mask.sum() == psi.sum_of_squares_constituents
                
                O_ij = np.einsum(
                    'kj,kab->ajb',
                    psi.endomorphism_basis()[:, i, nonnull_mask],
                    psi.endomorphism_basis(),
                )

                basis[p:p+irrep.size, j:j+psi.sum_of_squares_constituents, :] = O_ij
                column_mask[p:p+irrep.size] = nonnull_mask
                j += psi.sum_of_squares_constituents

            p += irrep.size
        
        if rho.sum_of_squares_constituents > 1:
        
            # tensorprod = np.einsum('ia,jb,kij->abk', basis[..., 0], basis[..., 0], rho.endomorphism_basis()/rho.sum_of_squares_constituents)
            # norm = (tensorprod**2).sum(axis=-1)
            # ortho = norm > eps
            
            endom_basis = (
                    rho_H.change_of_basis_inv[column_mask, :]
                  @ rho.endomorphism_basis()
                  @ rho_H.change_of_basis[:, column_mask]
            )
            ortho = (endom_basis**2).sum(0) > eps

            assert ortho.sum() == column_mask.sum() * rho.sum_of_squares_constituents, (ortho, column_mask.sum(), rho.sum_of_squares_constituents)

            n, dependencies = connected_components(csgraph=csr_matrix(ortho), directed=False, return_labels=True)
            
            # check Frobenius' Reciprocity
            assert n * rho.sum_of_squares_constituents == m_psi * psi.sum_of_squares_constituents,\
                (n, rho.sum_of_squares_constituents, m_psi, psi.sum_of_squares_constituents, rho, psi)

            mask = np.zeros((ortho.shape[0]), dtype=np.bool)

            for i in range(n):
                columns = np.nonzero(dependencies == i)[0]
                assert len(columns) == rho.sum_of_squares_constituents
                selected_column = columns[0]
                mask[selected_column] = 1

            assert mask.sum() == n
            
            basis = basis[:, mask, :]
            
            assert basis.shape[1] == n
            
        basis = np.einsum('oi,ijp->ojp', rho_H.change_of_basis, basis)

        return basis

    def dimension_basis(self, rho: Tuple, psi: Tuple) -> Tuple[int, int, int]:
        r"""
        
        Return the tuple :math:`(\text{dim}_\rho, m, \text{dim}_\psi)`, i.e. the shape of the array returned by
        :meth:`~escnn.group.HomSpace.basis`.

        Args:
            rho (IrreducibleRepresentation): an irrep of `G` (or its id)
            psi (IrreducibleRepresentation): an irrep of `H` (or its id)

        """
        rho = self.G.irrep(*rho)
        psi = self.H.irrep(*psi)

        # Computing this restriction every time can be very expensive.
        # Representation.restrict(id) keeps a cache of the representations, so the restriction needs to be computed only
        # the first time it is called
        rho_H = rho.restrict(self.sgid)

        m_psi = rho_H.multiplicity(psi.id)
        # m_psi = 0
        # for irrep in rho_H.irreps:
        #     if self.H.irrep(*irrep) == psi:
        #         m_psi += 1
        
        # Frobenius' Reciprocity theorem
        multiplicity = m_psi * psi.sum_of_squares_constituents / rho.sum_of_squares_constituents
        
        assert np.isclose(multiplicity, round(multiplicity))
        
        multiplicity = int(round(multiplicity))
        
        return rho.size, multiplicity, psi.size

    def scalar_basis(self,
              g: GroupElement,
              rho: Union[IrreducibleRepresentation, Tuple],
        ) -> np.ndarray:
        r"""

        Let `rho` be an irrep of :math:`G`.
        This method generates a basis for the subspace of scalar fields over :math:`X\cong G/H` which transforms
        under `rho` and samples its elements on the input :math:`g \in G`.

        .. note::
            Note that a scalar field :math:`f` is interpreted as a Mackey function, i.e. as a map
            :math:`f: G \to \R`, rather than :math:`f: X \to \R`.
            Indeed, this function takes an element :math:`g\in G` in input.
            Since this function is constant along each coset :math:`gH`, it can be composed with a choice of *section*
            :math:`\gamma: X \to G` to obtain a scalar field over `X`.

        Let :math:`m` be the multiplicity of :math:`\rho`, i.e. the number of subspaces which transform according to
        `rho` and :math:`\text{dim}_\rho` the dimensionality of the :math:`G`-irrep `rho`,
        Then, the space is :math:`\text{dim}_\rho \times m` dimensional and this method generates
        :math:`\text{dim}_\rho \times m` basis functions over :math:`G`.

        .. seealso::
            This method is equivalent to :meth:`escnn.group.HomSpace.basis` with ``psi = H.trivial_representation`` and
            flattening the last dimensionsion of the returned array (since the trivial representation is one
            dimensional).


        Args:
            g (GroupElement): the group element where to sample the elements of the basis
            rho (IrreducibleRepresentation): an irrep of `G` (or its id)

        Returns:
            an array of shape :math:`\text{dim}_\rho \times m` representing the basis elements
            sampled on `g`

        """

        return self.basis(g, rho, self.H.trivial_representation.id)

    def induced_representation(
            self,
            psi: Union[IrreducibleRepresentation, Tuple] = None,
            irreps: List = None,
            name: str = None
    ) -> Representation:
        r"""
            Representation acting on the finite dimensional invariant subspace of the induced representation containing
            only the ``irreps`` passed in input.
            The induced representation is expressed in the spectral basis, i.e. as a direct sum of irreps.

            The optional parameter ``name`` is also used for caching purpose.
            Consecutive calls of this method using the same ``name`` will ignore the arguments ``psi`` and ``irreps``
            and return the same instance of representation.


            .. note::

                If ``irreps`` does not contain sufficiently many irreps, the space might be 0-dimensional.
                In this case, this method returns None.

        """
        if name is None or name not in self._representations:

            if isinstance(psi, tuple):
                psi = self.H.irrep(*psi)
            assert isinstance(psi, IrreducibleRepresentation)
            assert psi.group == self.H

            assert irreps is not None and len(irreps) > 0, irreps

            _irreps = []
            for irr in irreps:
                if isinstance(irr, tuple):
                    irr = self.G.irrep(*irr)
                assert irr.group == self.G
                _irreps.append(irr.id)
            irreps = _irreps

            # check there are no duplicates
            assert len(irreps) == len(set(irreps)), irreps

        if name is None:
            irreps_names = '|'.join(str(i) for i in irreps)
            name = f'induced[{self.sgid}]_[{psi.id}]_[{irreps_names}]'

        if name not in self._representations:

            assert irreps is not None and len(irreps) > 0, irreps

            irreps_ids = []
            size = 0
            for irr in irreps:
                irr_size, multiplicity = self.dimension_basis(irr, psi.id)[:2]
                irreps_ids += [irr] * multiplicity

                size += multiplicity * irr_size

            if size == 0:
                return None

            self._names_to_irreps[name] = irreps
            self._names_to_psi[name] = psi.id

            supported_nonlinearities = ['norm', 'gated', 'concatenated']
            self._representations[name] = Representation(self.G,
                                                         name,
                                                         irreps_ids,
                                                         change_of_basis=np.eye(size),
                                                         supported_nonlinearities=supported_nonlinearities,
                                                         )

        return self._representations[name]

    def complete_basis(
            self,
            g: GroupElement,
            psi: Union[IrreducibleRepresentation, Tuple] = None,
            irreps: List = None,
            name: str = None
    ) -> Representation:

        r"""

        Let `psi` an irrep of :math:`H`.
        This method generates a basis for a subspace of `psi`-vector fields over :math:`X\cong G/H` and samples its
        elements on the input :math:`g \in G`.
        In particular, the method consider the union of all subspaces according to any irrep :math:`\rho` of :math:`G`
        in the input list ``irreps``.

        The parameters ``psi``, ``irreps`` and ``name`` are used to construct the corresponding representation by using
        the method :meth:`~escnn.group.HomSpace.induced_representation`.
        See that method's documentation.
        In particular, consecutive calls of this method with the same ``name`` parameter ignores the other two arguments
        and use the same cached representation.

        .. note::

            If ``irreps`` does not contain sufficiently many irreps, the space might be 0-dimensional.
            In this case, this method returns an empty array with shape 0.

        .. seealso::
            This method is equivalent to :meth:`escnn.group.HomSpace.basis`, called with ``rho`` being each irrep in
            ``irreps``. The resulting arrays are then properly reshaped and stacked.


        Args:
            g (GroupElement): the group element where to sample the elements of the basis
            psi (IrreducibleRepresentation): an irrep of `H` (or its id)
            irreps (list): a list of irreps of `G`
            name (str): name of the induced representation of `psi` (used for caching)

        Returns:
            an array of size equal to the size of the representation generated by
            :meth:`~escnn.group.HomSpace.induced_representation` using the same arguments.

        """

        ind_repr = self.induced_representation(psi, irreps, name)

        if ind_repr is None:
            return np.zeros(0)

        basis = np.empty(ind_repr.size)
        p = 0
        for rho in self._names_to_irreps[ind_repr.name]:

            basis_rho = self.basis(g, rho, self._names_to_psi[ind_repr.name])[:, :, 0]

            d = basis_rho.shape[0] * basis_rho.shape[1]

            basis[p:p+d] = basis_rho.T.reshape(-1)

            p += d

        return basis

    def _unit_test_basis(self):
        # check the equivariance of the generated basis

        for rho in self.G.irreps():
            rho_H = rho.restrict(self.sgid)

            for psi in self.H.irreps():

                for _ in range(30):
                    g1 = self.G.sample()
                    g2 = self.G.sample()

                    k_1 = self.basis(g1, rho, psi)
                    k_2 = self.basis(g2, rho, psi)

                    assert k_1.shape == self.dimension_basis(rho.id, psi.id)
                    assert k_2.shape == self.dimension_basis(rho.id, psi.id)

                    g12 = g2 @ (~g1)

                    assert np.allclose(
                        k_2,
                        np.einsum('oi, ijp->ojp', rho(g12), k_1)
                    )

                for _ in range(30):
                    h = self.H.sample()
                    g = self.G.sample()

                    B = self.basis(g, rho, psi)
                    assert B.shape == self.dimension_basis(rho.id, psi.id)

                    Bh = np.einsum('ijp,pq->ijq', B, psi(h))
                    hB = np.einsum('oi,ijp->ojp', rho(g) @ rho_H(h) @ rho(g).T, B)

                    assert np.allclose(
                        Bh, hB
                    )

            if self.H.order() == 1:
                # when inducing from the trivial group, one obtains the regular representation of G
                # (up to a permutation of the columns)

                for _ in range(100):
                    g = self.G.sample()

                    B = self.basis(g, rho, self.H.trivial_representation)

                    rho_g = rho(g)[:, :rho.size // rho.sum_of_squares_constituents]

                    # rho_g and B[..., 0] should be equal to each other up to a permutation of the columns
                    comparison = np.einsum('ij,ik->jk', rho_g, B[..., 0])

                    # therefore the comparison matrix needs to be a permutation matrix
                    assert (np.isclose(comparison.sum(axis=0), 1.)).all()
                    assert (np.isclose(comparison.sum(axis=1), 1.)).all()
                    assert (np.isclose(comparison, 0.) | np.isclose(comparison, 1.)).all()

    def _unit_test_full_basis(self):
        # check the equivariance of the generated basis

        irreps = []
        for rho in self.G.irreps():
            irreps.append(rho)

            for psi in self.H.irreps():

                ind_repr = self.induced_representation(psi, irreps)

                if ind_repr is None:
                    continue

                for _ in range(30):
                    g1 = self.G.sample()
                    g2 = self.G.sample()

                    k_1 = self.complete_basis(g1, psi, irreps, ind_repr.name)
                    k_2 = self.complete_basis(g2, psi, irreps, ind_repr.name)

                    g12 = g2 @ (~g1)

                    assert np.allclose(
                        k_2,
                        np.einsum('oi, i->o', ind_repr(g12), k_1)
                    )






from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from .equivariant_module import EquivariantModule

from typing import Tuple, Optional, Union

import torch
import numpy as np

import math

from torch.nn.functional import interpolate

__all__ = ["R2Upsampling"]


class R2Upsampling(EquivariantModule):
    
    def __init__(self,
                 in_type: FieldType,
                 scale_factor: Optional[int] = None,
                 size: Optional[Union[int, Tuple[int, int]]] = None,
                 mode: str = "bilinear",
                 align_corners: bool = False
                 ):
        r"""
        
        Wrapper for :func:`torch.nn.functional.interpolate`. Check its documentation for further details.
        
        Only ``"bilinear"`` and ``"nearest"`` methods are supported.
        However, ``"nearest"`` is not equivariant; using this method may result in broken equivariance.
        For this reason, we suggest to use ``"bilinear"`` (default value).

        .. warning ::
            The module supports a ``size`` parameter as an alternative to ``scale_factor``.
            However, the use of ``scale_factor`` should be *preferred*, since it guarantees both axes are scaled
            uniformly, which preserves rotation equivariance.
            A misuse of the parameter ``size`` can break the overall equivariance, since it might scale the two axes by
            two different factors.
        
        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.
        
        
        Args:
            in_type (FieldType): the input field type
            size (optional, int or tuple): output spatial size.
            scale_factor (optional, int): multiplier for spatial size
            mode (str): algorithm used for upsampling: ``nearest`` | ``bilinear``. Default: ``bilinear``
            align_corners (bool): if ``True``, the corner pixels of the input and output tensors are aligned, and thus
                    preserving the values at those pixels. This only has effect when mode is ``bilinear``.
                    Default: ``False``
            
        """

        assert isinstance(in_type.gspace, GSpace)
        assert in_type.gspace.dimensionality == 2

        super(R2Upsampling, self).__init__()

        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = in_type

        assert size is None or scale_factor is None, \
            f'Only one of "size" and "scale_factor" can be set, but found scale_factor={scale_factor} and size={size}'
        
        self._size = (size, size) if isinstance(size, int) else size
        assert self._size is None or (isinstance(self._size, tuple) and len(self._size) ==2), self._size
        self._scale_factor = scale_factor
        self._mode = mode
        self._align_corners = align_corners if mode != "nearest" else None
        
        if mode not in ["nearest", "bilinear"]:
            raise ValueError(f'Error Upsampling mode {mode} not recognized! Mode should be `nearest` or `bilinear`.')
        
    def forward(self, input: GeometricTensor):
        r"""
        
        Args:
            input (torch.Tensor): input feature map

        Returns:
             the result of the convolution
             
        """
        
        assert input.type == self.in_type
        assert len(input.shape) == 4

        if self._align_corners is None:
            output = interpolate(input.tensor,
                                 size=self._size,
                                 scale_factor=self._scale_factor,
                                 mode=self._mode)
        else:
            output = interpolate(input.tensor,
                                 size=self._size,
                                 scale_factor=self._scale_factor,
                                 mode=self._mode,
                                 align_corners=self._align_corners)
        
        return GeometricTensor(output, self.out_type, coords=None)

    def evaluate_output_shape(self, input_shape: Tuple[int, int, int, int]) -> Tuple[int, int, int, int]:
        assert len(input_shape) == 4
        assert input_shape[1] == self.in_type.size
    
        b, c, hi, wi = input_shape
        
        if self._size is None:
            ho = math.floor(hi * self._scale_factor)
            wo = math.floor(wi * self._scale_factor)
        else:
            ho = self._size[0]
            wo = self._size[1]

        return b, self.out_type.size, ho, wo
        
    def check_equivariance(self, atol: float = 0.1, rtol: float = 0.1):
        
        initial_size = 55

        c = self.in_type.size

        # x = torch.randn(3, c, initial_size, initial_size)

        import matplotlib.image as mpimg
        from skimage.transform import resize

        x = mpimg.imread('../group/testimage.jpeg').transpose((2, 0, 1))[np.newaxis, 0:c, :, :]
        x = x / 255.0
        x = resize(
            x,
            (x.shape[0], x.shape[1], initial_size, initial_size),
            anti_aliasing=True
        )

        if x.shape[1] < c:
            to_stack = [x for i in range(c // x.shape[1])]
            if c % x.shape[1] > 0:
                to_stack += [x[:, :(c % x.shape[1]), ...]]

            x = np.concatenate(to_stack, axis=1)

        x = GeometricTensor(torch.FloatTensor(x), self.in_type)

        errors = []

        for el in self.space.testing_elements:

            out1 = self(x).transform(el).tensor.detach().numpy()
            out2 = self(x.transform(el)).tensor.detach().numpy()

            b, c, h, w = out2.shape

            center_mask = np.zeros((2, h, w))
            center_mask[1, :, :] = np.arange(0, w) - w / 2
            center_mask[0, :, :] = np.arange(0, h) - h / 2
            center_mask[0, :, :] = center_mask[0, :, :].T
            center_mask = center_mask[0, :, :] ** 2 + center_mask[1, :, :] ** 2 < (h * 0.4) ** 2

            out1 = out1[..., center_mask]
            out2 = out2[..., center_mask]

            out1 = out1.reshape(-1)
            out2 = out2.reshape(-1)

            errs = np.abs(out1 - out2)

            esum = np.maximum(np.abs(out1), np.abs(out2))
            esum[esum == 0.0] = 1

            relerr = errs / esum

            # print(el, relerr.max(), relerr.mean(), relerr.var(), errs.max(), errs.mean(), errs.var())

            # tol = rtol*(np.abs(out1) + np.abs(out2)) + atol
            tol = rtol * esum + atol

            if np.any(errs > tol):
                print(el, relerr.max(), relerr.mean(), relerr.var(), errs.max(), errs.mean(), errs.var())
                # print(errs[errs > tol])
                print(out1[errs > tol])
                print(out2[errs > tol])
                print(tol[errs > tol])

            # assert np.all(np.abs(out1 - out2) < tol), 'The error found during equivariance check with element "{}" is too high: max = {}, mean = {} var ={}'.format(el, errs.max(), errs.mean(), errs.var())
            assert np.all(errs < tol), 'The error found during equivariance check with element "{}" is too high: max = {}, mean = {} var ={}'.format(el, errs.max(), errs.mean(), errs.var())

            errors.append((el, errs.mean()))

        return errors

    def export(self):
        r"""
        Export this module to a normal PyTorch :class:`torch.nn.Upsample` module and set to "eval" mode.

        """
    
        self.eval()
        
        if self._align_corners is not None:
            upsample = torch.nn.Upsample(
                size=self._size,
                scale_factor=self._scale_factor,
                mode=self._mode,
                align_corners=self._align_corners
            )
        else:
            upsample = torch.nn.Upsample(
                size=self._size,
                scale_factor=self._scale_factor,
                mode=self._mode,
            )
        
        return upsample.eval()



from .norm_max import NormMaxPool

from .pointwise_max import PointwiseMaxPool
from .pointwise_max import PointwiseMaxPoolAntialiased
from .pointwise_avg import PointwiseAvgPool
from .pointwise_avg import PointwiseAvgPoolAntialiased
from .pointwise_adaptive_avg import PointwiseAdaptiveAvgPool
from .pointwise_adaptive_max import PointwiseAdaptiveMaxPool

from .pointwise_avg_3d import PointwiseAvgPool3D
from .pointwise_avg_3d import PointwiseAvgPoolAntialiased3D


__all__ = [
    "NormMaxPool",
    "PointwiseMaxPool",
    "PointwiseMaxPoolAntialiased",
    "PointwiseAvgPool",
    "PointwiseAvgPoolAntialiased",
    "PointwiseAdaptiveAvgPool",
    "PointwiseAdaptiveMaxPool",
    ###### 3D ##################
    "PointwiseAvgPool3D",
    "PointwiseAvgPoolAntialiased3D",
]






from collections import defaultdict


from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from ..equivariant_module import EquivariantModule

import torch
import torch.nn.functional as F
from torch.nn import Parameter
from typing import List, Tuple, Any


__all__ = ["FieldDropout"]


def dropout_field(input: torch.Tensor, p: float, training: bool, inplace: bool):
    
    if training:
        shape = list(input.size())
        shape[2] = 1
        
        if input.device == torch.device('cpu'):
            mask = torch.FloatTensor(*shape)
        else:
            device = input.device
            mask = torch.cuda.FloatTensor(*shape, device=device)
        
        mask = mask.uniform_() > p
        mask = mask.to(torch.float)
        
        if inplace:
            input *= mask / (1. - p)
            return input
        else:
            return input * mask / (1. - p)
    else:
        return input


class FieldDropout(EquivariantModule):
    
    def __init__(self,
                 in_type: FieldType,
                 p: float = 0.5,
                 inplace: bool = False
                 ):
        r"""
        
        Applies dropout to individual *fields* independently.
        
        Notice that, with respect to :class:`~escnn.nn.PointwiseDropout`, this module acts on a whole field instead
        of single channels.
        
        Args:
            in_type (FieldType): the input field type
            p (float, optional): dropout probability
            inplace (bool, optional): can optionally do the operation in-place. Default: ``False``
            
        """

        assert isinstance(in_type.gspace, GSpace)
        if p < 0 or p > 1:
            raise ValueError("dropout probability has to be between 0 and 1, but got {}".format(p))
        
        super(FieldDropout, self).__init__()

        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = in_type
        
        self.p = p
        self.inplace = inplace
        
        self._nfields = None
        
        # group fields by their size and
        #   - check if fields of the same size are contiguous
        #   - retrieve the indices of the fields

        # number of fields of each size
        self._nfields = defaultdict(int)
        
        # indices of the channels corresponding to fields belonging to each group
        _indices = defaultdict(lambda: [])
        
        # whether each group of fields is contiguous or not
        self._contiguous = {}
        
        position = 0
        last_size = None
        for i, r in enumerate(self.in_type.representations):
            
            if r.size != last_size:
                if not r.size in self._contiguous:
                    self._contiguous[r.size] = True
                else:
                    self._contiguous[r.size] = False
            last_size = r.size
            
            _indices[r.size] += list(range(position, position + r.size))
            self._nfields[r.size] += 1
            position += r.size
        
        for s, contiguous in self._contiguous.items():
            if contiguous:
                # for contiguous fields, only the first and last indices are kept
                _indices[s] = torch.LongTensor([min(_indices[s]), max(_indices[s])+1])
            else:
                # otherwise, transform the list of indices into a tensor
                _indices[s] = torch.LongTensor(_indices[s])
                
            # register the indices tensors as parameters of this module
            self.register_buffer('indices_{}'.format(s), _indices[s])
        
        self._order = list(self._contiguous.keys())
        
    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""
        
        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map
            
        """
        
        assert input.type == self.in_type
        
        if not self.training:
            return input

        coords = input.coords
        input = input.tensor
        
        if not self.inplace:
            output = torch.empty_like(input)

        # iterate through all field sizes
        for s in self._order:
            
            indices = getattr(self, f"indices_{s}")
            
            shape = input.shape[:1] + (self._nfields[s], s) + input.shape[2:]
            
            if self._contiguous[s]:
                # if the fields are contiguous, we can use slicing
                out = dropout_field(input[:, indices[0]:indices[1], ...].view(shape), self.p, self.training, self.inplace)
                if not self.inplace:
                    shape = input.shape[:1] + (self._nfields[s] * s, ) + input.shape[2:]
                    output[:, indices[0]:indices[1], ...] = out.view(shape)
            else:
                # otherwise we have to use indexing
                out = dropout_field(input[:, indices, ...].view(shape), self.p, self.training, self.inplace)
                if not self.inplace:
                    shape = input.shape[:1] + (self._nfields[s] * s, ) + input.shape[2:]
                    output[:, indices, ...] = out.view(shape)
            
        if self.inplace:
            output = input
        
        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, coords)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:
        assert len(input_shape) > 1
        assert input_shape[1] == self.in_type.size
    
        return input_shape

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
        # return super(NormBatchNorm, self).check_equivariance(atol=atol, rtol=rtol)
        pass



from __future__ import annotations

import escnn.group
from escnn.group import Group, GroupElement
from escnn.group import IrreducibleRepresentation, Representation
from escnn.group import utils

from .utils import *

import numpy as np
import math

from typing import Tuple, Callable, Iterable, List, Dict, Any


__all__ = ["CyclicGroup"]


class CyclicGroup(Group):

    PARAM = 'int'
    PARAMETRIZATIONS = [
        'int',          # integer in 0, 1, ..., N-1
        'radians',      # real in 0., 2pi/N, ... i*2pi/N, ...
        # 'C',            # point in the unit circle (i.e. cos and sin of 'radians')
        'MAT',          # 2x2 rotation matrix
    ]

    def __init__(self, N: int):
        r"""
        Build an instance of the cyclic group :math:`C_N` which contains :math:`N` discrete planar rotations.
        
        The group elements are :math:`\{e, r, r^2, r^3, \dots, r^{N-1}\}`, with group law
        :math:`r^a \cdot r^b = r^{\ a + b \!\! \mod \!\! N \ }`.
        The cyclic group :math:`C_N` is isomorphic to the integers *modulo* ``N``.
        For this reason, elements are stored as the integers between :math:`0` and :math:`N-1`, where the :math:`k`-th
        element can also be interpreted as the discrete rotation by :math:`k\frac{2\pi}{N}`.
        
        Subgroup Structure.
        
        A subgroup of :math:`C_N` is another cyclic group :math:`C_M` and is identified by an ``id`` containing the
        integer :math:`M` (i.e. the order of the subgroup).
        
        If the current group is :math:`C_N`, the subgroup is generated by :math:`r^{(N/M)}`.
        Notice that :math:`M` has to divide the order :math:`N` of the group.
        
        Args:
            N (int): order of the group
            
        """
        
        assert (isinstance(N, int) and N > 0), N
        
        super(CyclicGroup, self).__init__("C%d" % N, False, True)
        
        self.N = N
        
        self._elements = [self.element(i) for i in range(N)]
        # self._elements_names = ['e'] + ['r%d' % i for i in range(1, N)]

        self._identity = self.element(0)
        
        self._build_representations()

    @property
    def generators(self) -> List[GroupElement]:
        if self.order() > 1:
            return [self.element(1)]
        else:
            # the generator of the trivial group is the empty set
            return []

    @property
    def identity(self) -> GroupElement:
        return self._identity

    @property
    def elements(self) -> List[GroupElement]:
        return self._elements

    # @property
    # def elements_names(self) -> List[str]:
    #     return self._elements_names

    @property
    def _keys(self) -> Dict[str, Any]:
        return {'N': self.N}

    @property
    def subgroup_trivial_id(self):
        return 1

    @property
    def subgroup_self_id(self):
        return self.order()

    ###########################################################################
    # METHODS DEFINING THE GROUP LAW AND THE OPERATIONS ON THE GROUP'S ELEMENTS
    ###########################################################################

    def _inverse(self, element: int, param: str = PARAM) -> int:
        r"""
        Return the inverse element :math:`r^{-j \mod N}` of the input element :math:`r^j`, specified by the input
        integer :math:`j` (``element``)
        
        Args:
            element (int): a group element :math:`r^j`

        Returns:
            its opposite :math:`r^{-j \mod N}`
            
        """
        self._change_param(element, param, 'int')
        element = (-element) % self.N
        return self._change_param(element, 'int', param)

    def _combine(self, e1: int, e2: int, param: str = PARAM, param1: str = None, param2: str = None) -> int:
        r"""
        Return the composition of the two input elements.
        Given two integers :math:`a` and :math:`b` representing the elements :math:`r^a` and :math:`r^b`, the method
        returns the integer :math:`a + b \mod N` representing the element :math:`r^{a + b \mod N}`.
        

        Args:
            e1 (int): a group element :math:`r^a`
            e2 (int): another group element :math:`r^a`

        Returns:
            their composition :math:`r^{a+b \mod N}`
            
        """

        if param1 is None:
            param1 = param
        if param2 is None:
            param2 = param

        e1 = self._change_param(e1, p_from=param1, p_to='int')
        e2 = self._change_param(e2, p_from=param2, p_to='int')
        return self._change_param(
            (e1 + e2) % self.N,
            p_from = 'int',
            p_to = param
        )

    def _equal(self, e1: int, e2: int, param: str = PARAM, param1: str = None, param2: str = None) -> bool:
        r"""

        Check if the two input values corresponds to the same element.

        Args:
            e1 (int): an element
            e2 (int): another element

        Returns:
            whether they are the same element

        """
        if param1 is None:
            param1 = param
        if param2 is None:
            param2 = param

        e1 = self._change_param(e1, p_from=param1, p_to='int')
        e2 = self._change_param(e2, p_from=param2, p_to='int')
        return e1 == e2
    
    def _is_element(self, element: int, param: str = PARAM, verbose: bool = False) -> bool:
    
        element = self._change_param(element, p_from=param, p_to='int')
        if isinstance(element, int):
            return 0 <= element < self.N
        else:
            return False

    def _hash_element(self, element: int, param: str = PARAM):
        element = self._change_param(element, p_from=param, p_to='int')
        return hash(element)

    def _repr_element(self, element: int, param: str = PARAM):
        element = self._change_param(element, p_from=param, p_to='int')
        return "{}[2pi/{}]".format(element, self.N)
    
    def _change_param(self, element, p_from: str, p_to: str):
        assert p_from in self.PARAMETRIZATIONS
        assert p_to in self.PARAMETRIZATIONS
        
        if p_from == 'MAT':
            assert isinstance(element, np.ndarray)
            assert element.shape == (2, 2)
            assert np.isclose(np.linalg.det(element), 1.)
            assert np.allclose(element @ element.T, np.eye(2))
    
            cos = (element[0, 0] + element[1, 1]) / 2.
            sin = (element[1, 0] - element[1, 0]) / 2.
    
            element = np.arctan2(sin, cos)
            p_from = 'radians'

        # convert to INT
        if p_from == 'int':
            assert isinstance(element, int), element
        elif p_from == 'radians':
            assert isinstance(element, float), element
            if not utils.cycle_isclose(element, 0., 2*np.pi/self.N):
                raise ValueError()
            element = int(round(self.N * element / (2*np.pi))) % self.N
        else:
            raise ValueError('Parametrization {} not recognized'.format(p_from))

        # convert from INT
        if p_to == 'int':
            return element
        elif p_to == 'radians':
            return element * (2*np.pi) / self.N
        elif p_to == 'MAT':
            element = element * (2*np.pi) / self.N
            cos = np.cos(element)
            sin = np.sin(element)
            return np.array(([
                [cos, -sin],
                [sin, cos],
            ]))
        else:
            raise ValueError('Parametrization {} not recognized'.format(p_to))

    ###########################################################################

    def sample(self) -> GroupElement:
        return self.element(np.random.randint(0, self.order()))

    def testing_elements(self) -> Iterable[GroupElement]:
        r"""
        A finite number of group elements to use for testing.
        
        """
        return iter(self._elements)

    def __eq__(self, other):
        if not isinstance(other, CyclicGroup):
            return False
        else:
            return self.name == other.name and self.order() == other.order()

    def _subgroup(self, id: int) -> Tuple[
        Group,
        Callable[[GroupElement], GroupElement],
        Callable[[GroupElement], GroupElement]
    ]:
        r"""
        Restrict the current group to the cyclic subgroup :math:`C_M`.
        If the current group is :math:`C_N`, it restricts to the subgroup generated by :math:`r^{(N/M)}`.
        Notice that :math:`M` has to divide the order :math:`N` of the current group.
        
        The method takes as input the integer :math:`M` identifying of the subgroup to build (the order of the subgroup)
        
        Args:
            id (int): the integer :math:`M` identifying of the subgroup

        Returns:
            a tuple containing

                - the subgroup,

                - a function which maps an element of the subgroup to its inclusion in the original group and

                - a function which maps an element of the original group to the corresponding element in the subgroup (returns None if the element is not contained in the subgroup)
                
        """

        assert isinstance(id, int), id

        order = id

        assert self.order() % order == 0, \
            "Error! The subgroups of a cyclic group have an order that divides the order of the supergroup." \
            " %d does not divide %d " % (order, self.order())

        # Build the subgroup
        
        # take the elements of the group generated by "r^ratio"
        sg = escnn.group.cyclic_group(order)

        # parent_mapping = lambda e, ratio=ratio: self.element(e._element * ratio)
        # child_mapping = lambda e, ratio=ratio, sg=sg: None if e._element % ratio != 0 else sg.element(int(e._element // ratio))
        parent_mapping = _build_parent_map(self, order)
        child_mapping = _build_child_map(self, sg)

        return sg, parent_mapping, child_mapping

    def grid(self, type: str, N: int) -> List[GroupElement]:
        r"""
            .. todo ::
                Add docs

        """

        if type == 'rand':
            return [self.sample() for _ in range(N)]
        elif type == 'regular':
            assert self.order() % N == 0
            r = self.order() // N
            return [self.element(i*r) for i in range(N)]
        else:
            raise ValueError(f'Grid type "{type}" not recognized!')

    def _combine_subgroups(self, sg_id1, sg_id2):
    
        sg_id1 = self._process_subgroup_id(sg_id1)
        sg1, inclusion, restriction = self.subgroup(sg_id1)
        sg_id2 = sg1._process_subgroup_id(sg_id2)
        
        return sg_id2

    def _restrict_irrep(self, irrep: Tuple, id: int) -> Tuple[np.matrix, List[Tuple]]:
        r"""
        
        Restrict the input irrep to the subgroup :math:`C_m` with order ``m``.
        If the current group is :math:`C_n`, it restricts to the subgroup generated by :math:`r^{(n/m)}`.
        Notice that :math:`m` has to divide the order :math:`n` of the current group.
        
        The method takes as input the integer :math:`m` identifying of the subgroup to build (the order of the subgroup)

        Args:
            irrep (tuple): the identifier of the irrep to restrict
            id (int): the integer ``m`` identifying the subgroup

        Returns:
            a pair containing the change of basis and the list of irreps of the subgroup which appear in the restricted irrep
            
        """
    
        irr = self.irrep(*irrep)
    
        # Build the subgroup
        sg, _, _ = self.subgroup(id)
    
        order = id
    
        change_of_basis = None
        irreps = []
    
        f = irr.attributes["frequency"] % order
    
        if f > order/2:
            f = order - f
            change_of_basis = np.array([[1, 0], [0, -1]])
        else:
            change_of_basis = np.eye(irr.size)
    
        r = (f,)
    
        irreps.append(r)
        if sg.irrep(*r).size < irr.size:
            irreps.append(r)
        
        return change_of_basis, irreps

    def _build_representations(self):
        r"""
        Build the irreps and the regular representation for this group
        
        """
        
        N = self.order()

        # Build all the Irreducible Representations
        for k in range(0, int(N // 2) + 1):
            self.irrep(k)
            
        # Build all Representations

        # add all the irreps to the set of representations already built for this group
        self.representations.update(**{irr.name : irr for irr in self.irreps()})

        # build the regular representation
        self.representations['regular'] = self.regular_representation
        self.representations['regular'].supported_nonlinearities.add('vectorfield')
        
    def _build_quotient_representations(self):
        r"""
        Build all the quotient representations for this group

        """
        for n in range(2, int(math.ceil(math.sqrt(self.order())))):
            if self.order() % n == 0:
                self.quotient_representation(n)
    
    @property
    def trivial_representation(self) -> Representation:
        return self.representations['irrep_0']

    def irrep(self, k: int) -> IrreducibleRepresentation:
        r"""
        Build the irrep of frequency ``k`` of the current cyclic group.
        The frequency has to be a non-negative integer in :math:`\{0, \dots, \left \lfloor N/2 \right \rfloor \}`,
        where :math:`N` is the order of the group.
        
        Args:
            k (int): the frequency of the representation

        Returns:
            the corresponding irrep

        """
        id = (k,)
        
        if id not in self._irreps:
            
            assert 0 <= k <= self.order() // 2, (k, self.order())
            name = f"irrep_{k}"
            
            n = self.order()
            
            if k == 0:
                # Trivial representation
            
                irrep = _build_irrep_cn(0)
                character = _build_char_cn(0)
                supported_nonlinearities = ['pointwise', 'gate', 'norm', 'gated', 'concatenated']
                self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, 1, 'R',
                                                            supported_nonlinearities=supported_nonlinearities,
                                                            character=character,
                                                            # trivial=True,
                                                            frequency=k)
            elif n % 2 == 0 and k == int(n/2):
                # 1 dimensional Irreducible representation (only for even order groups)
                irrep = _build_irrep_cn(k)
                character = _build_char_cn(k)
                supported_nonlinearities = ['norm', 'gated', 'concatenated']
                self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, 1, 'R',
                                                            supported_nonlinearities=supported_nonlinearities,
                                                            character=character,
                                                            frequency=k)
            else:
                # 2 dimensional Irreducible Representations

                irrep = _build_irrep_cn(k)
                character = _build_char_cn(k)

                supported_nonlinearities = ['norm', 'gated']
                self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, 2, 'C',
                                                            supported_nonlinearities=supported_nonlinearities,
                                                            character=character,
                                                            frequency=k)
        return self._irreps[id]

    def _clebsh_gordan_coeff(self, m, n, j) -> np.ndarray:
        m, = self.get_irrep_id(m)
        n, = self.get_irrep_id(n)
        j, = self.get_irrep_id(j)

        rho_m = self.irrep(m)
        rho_n = self.irrep(n)
        rho_j = self.irrep(j)

        if m == 0 or n == 0:
            if j == m + n:
                return np.eye(rho_j.size).reshape(rho_m.size, rho_n.size, 1, rho_j.size)
            else:
                return np.zeros((rho_m.size, rho_n.size, 0, rho_j.size))
        elif (self.N % 2 == 0) and (m == self.N//2 or n == self.N//2):
            if j == m + n:
                return np.eye(rho_j.size).reshape(rho_m.size, rho_n.size, 1, rho_j.size)
            elif j == (self.N -m - n):
                cg = np.eye(rho_j.size)
                if rho_j.size > 1:
                    cg[:, 1] *= -1
                return cg.reshape(rho_m.size, rho_n.size, 1, rho_j.size)
            else:
                return np.zeros((rho_m.size, rho_n.size, 0, rho_j.size))
        else:
            cg = np.array([
                [1., 0., 1., 0.],
                [0., 1., 0., 1.],
                [0., -1., 0., 1.],
                [1., 0., -1., 0.],
            ]) / np.sqrt(2)
            if j == m + n:
                cg = cg[:, 2:]
            elif j == self.N - m - n:
                cg = cg[:, 2:]
                cg[:, 1] *= -1
            elif j == m - n:
                cg = cg[:, :2]
            elif j == n - m:
                cg = cg[:, :2]
                cg[:, 1] *= -1
            else:
                cg = np.zeros((rho_m.size, rho_n.size, 0, rho_j.size))

            return cg.reshape(rho_n.size, rho_m.size, -1, rho_j.size).transpose(1, 0, 2, 3)

    def _tensor_product_irreps(self, J: int, l: int) -> List[Tuple[Tuple, int]]:
        J, = self.get_irrep_id(J)
        l, = self.get_irrep_id(l)
    
        if J == 0 or l == 0:
            return [
                ((l + J,), 1)
            ]
        elif (self.N % 2 == 0) and (J == self.N // 2 or l == self.N // 2):
            j = (J + l) if (J+l <= self.N//2) else (self.N - J - l)
            return [
                ((j,), 1)
            ]
        elif l == J:
            j = (J + l) if (J+l <= self.N//2) else (self.N - J - l)
            m = 1 if j < self.N/2 else 2
            return [
                ((0,), 2),
                ((j,), m),
            ]
        else:
            j = (J + l) if (J+l <= self.N//2) else (self.N - J - l)
            m = 1 if j < self.N/2 else 2
            return [
                ((np.abs(l - J),), 1),
                ((j,), m),
            ]

    _cached_group_instances = {}
    
    @classmethod
    def _generator(cls, N: int) -> 'CyclicGroup':
        if N not in cls._cached_group_instances:
            cls._cached_group_instances[N] = CyclicGroup(N)
        
        return cls._cached_group_instances[N]


def _build_irrep_cn(k: int):
    def irrep(element: GroupElement, k:int =k) -> np.ndarray:
        if k == 0:
            return np.eye(1)
        
        n = element.group.order()

        if n % 2 == 0 and k == int(n / 2):
            # 1 dimensional Irreducible representation (only for even order groups)
            return np.array([[np.cos(k * element.to('radians'))]])
        else:
            # 2 dimensional Irreducible Representations
            return utils.psi(element.to('radians'), k=k)
        
    return irrep


def _build_char_cn(k: int):
    
    def character(element: GroupElement, k=k) -> float:
        if k == 0:
            return 1.
        
        n = element.group.order()
        
        if n % 2 == 0 and k == int(n / 2):
            # 1 dimensional Irreducible representation (only for even order groups)
            return np.cos(k * element.to('radians'))
        else:
            # 2 dimensional Irreducible Representations
            return 2*np.cos(k * element.to('radians'))
    
    return character


def _build_parent_map(G: CyclicGroup, order: int):
    def parent_mapping(e: GroupElement, G: Group = G, order=order) -> GroupElement:
        return G.element(e.to('int') * G.order() // order)
    
    return parent_mapping


def _build_child_map(G: CyclicGroup, sg: CyclicGroup):
    assert G.order() % sg.order() == 0
    
    def child_mapping(e: GroupElement, G=G, sg: Group = sg) -> GroupElement:
        assert e.group == G
        i = e.to('int')
        ratio = G.order() // sg.order()
        if i % ratio != 0:
            return None
        else:
            return sg.element(i // ratio)
    
    return child_mapping


